# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click Command definition for the class command group which includes
commands for get, enumerate, associators, references, find, etc. of the objects
CIMClass on a WBEM server.

NOTE: Commands are ordered in help display by their order in this file.
"""

from __future__ import absolute_import, print_function

from collections import namedtuple
import click

from pywbem._nocasedict import NocaseDict

from pywbem import Error, CIMClassName, CIMError, ModelError, CIM_ERR_NOT_FOUND

from .pywbemcli import cli
from ._common import filter_namelist, \
    resolve_propertylist, CMD_OPTS_TXT, GENERAL_OPTS_TXT, SUBCMD_HELP_TXT, \
    output_format_is_table, format_table, process_invokemethod, \
    pywbem_error_exception, warning_msg, validate_output_format, \
    get_subclass_names, depending_classnames, get_leafclass_names

from ._display_cimobjects import display_cim_objects

from ._common_options import add_options, propertylist_option, \
    names_only_option, include_classorigin_class_option, namespace_option,  \
    summary_option, multiple_namespaces_option, class_filter_options, \
    help_option
from ._displaytree import display_class_tree
from ._click_extensions import PywbemcliGroup, PywbemcliCommand
from ._utils import pywbemcliwarn

#
#   Common option definitions for class group
#

# NOTE: A number of the options use double-dash as the short form.  In those
# cases, a third definition of the options without the double-dash defines
# the corresponding option name, ex. 'include_qualifiers'. It should be
# defined with underscore and not dash

no_qualifiers_class_option = [              # pylint: disable=invalid-name
    click.option('--nq', '--no-qualifiers', 'no_qualifiers', is_flag=True,
                 default=True,
                 help=u'Do not include qualifiers in the returned class(es). '
                      'Default: Include qualifiers.')]

deep_inheritance_class_option = [              # pylint: disable=invalid-name
    click.option('--di', '--deep-inheritance', 'deep_inheritance', is_flag=True,
                 default=False,
                 help=u'Include the complete subclass hierarchy of the '
                      'requested classes in the result set. '
                      'Default: Do not include subclasses.')]

local_only_class_option = [              # pylint: disable=invalid-name
    click.option('--lo', '--local-only', 'local_only', is_flag=True,
                 default=False,
                 help=u'Do not include superclass properties and methods in '
                      u'the returned class(es). '
                      u'Default: Include superclass properties and methods.')]


##########################################################################
#
#   Click command group and command definitions
#   These decorated functions implement the commands, arguments, and
#   options for the top-level class command group
#
###########################################################################

@cli.group('class', cls=PywbemcliGroup, options_metavar=GENERAL_OPTS_TXT,
           subcommand_metavar=SUBCMD_HELP_TXT)
@add_options(help_option)
def class_group():
    """
    Command group for CIM classes.

    This command group defines commands to inspect classes, invoke
    methods on classes, delete classes.

    Creation and modification of classes is not currently supported.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'class' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@class_group.command('enumerate', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=False)
@add_options(deep_inheritance_class_option)
@add_options(local_only_class_option)
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(class_filter_options)
@add_options(help_option)
@click.pass_obj
def class_enumerate(context, classname, **options):
    """
    List top classes or subclasses of a class in a namespace.

    Enumerate CIM classes starting either at the top of the class hierarchy
    in the specified CIM namespace (--namespace option), or at the specified
    class (CLASSNAME argument) in the specified namespace. If no namespace was
    specified, the default namespace of the connection is used.

    The --local-only, --include-classorigin, and --no-qualifiers options
    determine which parts are included in each retrieved class.

    The --deep-inheritance option defines whether or not the complete subclass
    hierarchy of the classes is retrieved.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class enumerate -n interop

      pywbemcli -n myconn class enumerate CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_enumerate(context, classname,
                                                    options))


@class_group.command('get', cls=PywbemcliCommand, options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@add_options(local_only_class_option)
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def class_get(context, classname, **options):
    """
    Get a class.

    Get a CIM class (CLASSNAME argument) in a CIM namespace (--namespace
    option). If no namespace was specified, the default namespace of the
    connection is used.

    The --local-only, --include-classorigin, --no-qualifiers, and
    --propertylist options determine which parts are included in each retrieved
    class.

    In the output, the class will be formatted as defined by the
    --output-format general option. Table formats are replaced with MOF
    format.

    Example:

      pywbemcli -n myconn class get CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_get(context, classname, options))


@class_group.command('delete', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@click.option('-f', '--force', is_flag=True, default=False,
              help=u'Same as --include-instances. The -f / --force option has '
                   'been deprecated and will be removed in a future version.')
@click.option('--include-instances', is_flag=True, default=False,
              help=u'Delete any instances of the class as well. '
                   'WARNING: Deletion of instances will cause the removal '
                   'of corresponding resources in the managed environment '
                   '(i.e. in the real world).'
                   'Default: Reject command if the class has any instances.')
@click.option('--dry-run', is_flag=True, required=False,
              help=u'Enable dry-run mode: Do not actually delete the objects, '
                   'but display what would be done.')
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def class_delete(context, classname, **options):
    """
    Delete a class.

    Delete a CIM class (CLASSNAME argument) in a CIM namespace (--namespace
    option). If no namespace was specified, the default namespace of the
    connection is used.

    If the class has instances, the command is rejected, unless the
    --include-instances option is specified, in which case the instances
    are also deleted.

    If other classes in that namespace depend on the class to be deleted, the
    command is rejected. Dependencies considered for this purpose are
    subclasses, referencing classes and embedding classes (EmbeddedInstance
    qualifier only).

    WARNING: Deletion of instances will cause the removal of corresponding
    resources in the managed environment (i.e. in the real world). Some
    instances may not be deletable.

    WARNING: Deleting classes can cause damage to the server: It can impact
    instance providers and other components in the server. Use this
    command with caution.

    Many WBEM servers may not allow this operation or may severely limit the
    conditions under which a class can be deleted from the server.

    Example:

      pywbemcli -n myconn class delete CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_delete(context, classname, options))


@class_group.command('invokemethod', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@click.argument('methodname', type=str, metavar='METHODNAME', required=True)
@click.option('-p', '--parameter', type=str, metavar='PARAMETERNAME=VALUE',
              required=False, multiple=True,
              help=u'Specify a method input parameter with its value. '
                   'May be specified multiple times. '
                   'Default: No input parameters.')
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def class_invokemethod(context, classname, methodname, **options):
    """
    Invoke a method on a class.

    Invoke a static CIM method (METHODNAME argument) on a CIM class (CLASSNAME
    argument) in a CIM namespace (--namespace option), and display the method
    return value and output parameters. If no namespace was specified, the
    default namespace of the connection is used.

    The method input parameters are specified using the --parameter option,
    which may be specified multiple times.

    Pywbemcli retrieves the class definition from the server in order to
    verify that the specified input parameters are consistent with the
    parameter characteristics in the method definition.

    Use the 'instance invokemethod' command to invoke CIM methods on CIM
    instances.

    Example:

      pywbemcli -n myconn class invokemethod CIM_Foo methodx -p p1=9 -p p2=Fred
    """
    context.execute_cmd(lambda: cmd_class_invokemethod(context, classname,
                                                       methodname, options))


@class_group.command('references', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help=u'Filter the result set by result class name. '
                   'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by source end role name.')
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(help_option)
@click.pass_obj
def class_references(context, classname, **options):
    """
    List the classes referencing a class.

    List the CIM (association) classes that reference the specified class
    (CLASSNAME argument) in the specified CIM namespace
    (--namespace option). If no namespace was specified, the default namespace
    of the connection is used.

    The classes to be retrieved can be filtered by the --role and
    --result-class options.

    The --include-classorigin, --no-qualifiers, and --propertylist options
    determine which parts are included in each retrieved class.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class references CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_references(context, classname,
                                                     options))


@class_group.command('associators', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@click.option('--ac', '--assoc-class', 'assoc_class', type=str, required=False,
              metavar='CLASSNAME',
              help=u'Filter the result set by association class name. '
                   'Subclasses of the specified class also match.')
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help=u'Filter the result set by result class name. '
                   'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by source end role name.')
@click.option('--rr', '--result-role', 'result_role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by far end role name.')
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(help_option)
@click.pass_obj
def class_associators(context, classname, **options):
    """
    List the classes associated with a class.

    List the CIM classes that are associated with the specified class
    (CLASSNAME argument) in the specified CIM namespace
    (--namespace option). If no namespace was specified, the default namespace
    of the connection is used.

    The classes to be retrieved can be filtered by the --role, --result-role,
    --assoc-class, and --result-class options.

    The --include-classorigin, --no-qualifiers, and --propertylist options
    determine which parts are included in each retrieved class.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class associators CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_associators(context, classname,
                                                      options))


@class_group.command('find', cls=PywbemcliCommand, options_metavar=CMD_OPTS_TXT)
@click.argument('classname-glob', type=str, metavar='CLASSNAME-GLOB',
                required=True)
@add_options(multiple_namespaces_option)
@click.option('-s', '--sort', is_flag=True, required=False,
              help=u'Sort by namespace. Default is to sort by classname')
@add_options(class_filter_options)
@add_options(help_option)
@click.pass_obj
def class_find(context, classname_glob, **options):
    """
    List the classes with matching class names on the server.

    Find the CIM classes whose class name matches the specified wildcard
    expression (CLASSNAME-GLOB argument) in all CIM namespaces of the
    WBEM server, or in the specified namespace (--namespace option).

    The CLASSNAME-GLOB argument is a wildcard expression that is matched on
    class names case insensitively.
    The special characters from Unix file name wildcarding are supported
    ('*' to match zero or more characters, '?' to match a single character,
    and '[]' to match character ranges). To avoid shell expansion of wildcard
    characters, the CLASSNAME-GLOB argument should be put in quotes.

    For example, "pywbem_*" returns classes whose name begins with "PyWBEM_",
    "pywbem_", etc. "*system*" returns classes whose names include the case
    insensitive string "system".

    In the output, the classes will be formatted as defined by the
    --output-format general option if it specifies table output. Otherwise
    the classes will be in the form "NAMESPACE:CLASSNAME".

    Examples:

      pywbemcli -n myconn class find "CIM_*System*" -n interop

      pywbemcli -n myconn class find *Foo*
    """
    context.execute_cmd(lambda: cmd_class_find(context, classname_glob,
                                               options))


@class_group.command('tree', cls=PywbemcliCommand, options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=False)
@click.option('-s', '--superclasses', is_flag=True, default=False,
              help=u'Show the superclass hierarchy. '
                   'Default: Show the subclass hierarchy.')
@click.option('-d', '--detail', is_flag=True, default=False,
              help=u'Show details about the class: the Version, '
                   ' Association, Indication, and Abstact qualifiers.')
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def class_tree(context, classname, **options):
    """
    Show the subclass or superclass hierarchy for a class.

    List the subclass or superclass hierarchy of a CIM class (CLASSNAME
    argument) or CIM namespace (--namespace option):

    - If CLASSNAME is omitted, the complete class hierarchy of the specified
      namespace is retrieved.

    - If CLASSNAME is specified but not --superclasses, the class and its
      subclass hierarchy in the specified namespace are retrieved.

    - If CLASSNAME and --superclasses are specified, the class and its
      superclass ancestry up to the top-level class in the specified namespace
      are retrieved.

    If no namespace was specified, the default namespace of the connection is
    used.

    The class hierarchy will formatted as a ASCII graphical tree; the
    --output-format general option is ignored.

    The --detail options to display extra information about each class
    including:

    -  The Version qualifier value if the class includes a version
       qualifier. This is normally a string with 3 integers

    -  Information about each class type (Association, Indication, Abstract)

    Examples:

      # Display the complete class hierarchy from the interop namespace
      pywbemcli -n myconn class tree -n interop

      # Display CIM_Foo an its subclasses from the namespace interop
      pywbemcli -n myconn class tree CIM_Foo -n interop

      # Display CIM_Foo and its superclasses from interop
      pywbemcli -n myconn class tree CIM_Foo -s -n interop

    """
    context.execute_cmd(lambda: cmd_class_tree(context, classname, options))


####################################################################
#
#  Common functions for cmd_class processing
#  This includes functions used by the command action functions
#  in this fmodule and possibly other modules
#
####################################################################

def parse_version_str(version_str):
    """
    Parse a string with 3 positive integers seperated by period (CIM version
    string) into a 3 integer tuple and return the tuole

    Parameters:
        version_str (:term: str):
            String defining 3 components of a CIM version

    Returns:
        tuple containing 3 integers
    """
    try:
        version_tuple = [int(x) for x in version_str.split('.')]
    except ValueError:
        raise click.ClickException('--since option value invalid. '
                                   'Must contain 3 integer elements: '
                                   'int.int.int". {} received'.
                                   format(version_str))
    if len(version_tuple) != 3:
        raise click.ClickException('Version value must contain 3 integer '
                                   'elements (int.int.int). '
                                   '{} received'.format(version_str))
    return version_tuple


# Namedtupe defining each entry in the filters dictionary values
FILTERDEF = namedtuple('FILTERDEF', 'optionvalue qualifiername scopes')


def _build_filters_dict(conn, ns, options):
    """
    Build a dictionary defining the filters to be processed against list
    of classes from
    the filter definitons in the Click options dictionary. There is an entry
    in the dictionary for each filter to be applied to filter a list of
    classes.

    Returns:
      Dict of filters where the names are the  filtersthemselves , the types
      are the type of test (i.e. qualifier, superclass)
      and the value for each is a tuple:
      * Name of the qualifier (:term:`string`)
      * The value of the qualifier filter option (True or False which
        determines whether to display the existence or non-existence of the
        qualifier
      * A tuple containing Booleans for the value of each of the possible
        element scopes (class, property, method, parameter) indicating whether
        the qualifier is allowed in that element.
    """
    filters = {}

    def set_qualifier_option(qname, option_value):
        qualdecl = conn.GetQualifier(qname, ns)
        # Note: qualdecl.scopes performs test case-insensitively
        if qualdecl.scopes['any']:
            scopes_map = [True, True, True, True]
        else:
            scopes_map = [False, False, False, False]
            scopes_map[0] = any([qualdecl.scopes['class'],
                                 qualdecl.scopes['association'],
                                 qualdecl.scopes['indication']])
            scopes_map[1] = qualdecl.scopes['property']
            scopes_map[2] = qualdecl.scopes['method']
            scopes_map[3] = qualdecl.scopes['parameter']
        filters['qualifier'] = FILTERDEF(option_value, qname,
                                         tuple(scopes_map))

    # Qualifier options
    if options['association'] is not None:
        set_qualifier_option('association', options['association'])
    if options['indication'] is not None:
        set_qualifier_option('indication', options['indication'])
    if options['experimental'] is not None:
        set_qualifier_option('experimental', options['experimental'])
    # If set, the entity is deprecated
    if options['deprecated'] is not None:
        set_qualifier_option('deprecated', options['deprecated'])
    if options['since'] is not None:
        version_tuple = parse_version_str(options['since'])
        set_qualifier_option('version', version_tuple)

    if options['schema'] is not None:
        test_str = "{}_".format(options['schema'].lower())
        filters['schema'] = FILTERDEF(test_str, None, None)

    if options['subclass_of'] is not None:
        filters['subclass_of'] = FILTERDEF(options['subclass_of'], None, None)
    if options['leaf_classes'] is not None:
        filters['leaf_classes'] = FILTERDEF(options['leaf_classes'], None, None)
    return filters


def _filter_classes(classes, filters, names_only, iq):
    """
    Filter a list of classes for the qualifiers defined by  the
    qualifier_filter parameter where this parameter is a list of tuples.
    each tuple contains the qualifier name and a dictionary with qualifier
     name as key and tuple containing the option_value(True or False) and
    a list of booleans where each boolean represents one of the scope types
    ()
    whether to display or not display if it exists.

    This method only works for boolean qualifiers

    Parameters:

      classes (list of :class:`~pywbem.CIMClass`):
        list of classes to be filtered

      qualifier_filters (dict):
        Dictionary defining the filtering to be performed. It contains an entry
        for each qualifier filter that is defined. See _build_qualifier_filters
        for a definition of this list.

      names_only (:class:`py:bool`):
        If True, return only the classnames. Otherwise returns the filtered
        classes. This is because we must get the classes from the server to
        perform the filtering

      iq (:class:`py:bool`):
        If not True, remove any qualifiers from the classes.  This is because
        we must get the classes from the server with qualifiers to
        perform the filtering.
    """

    def class_has_qualifier(cls, qname, scopes):
        """
        Determine if the qualifier defined by qname exists in the elements
        of the class where the elements are defined by the scopes parameter
        for this filter.

        Parameters:

          cls (:class:`~pywbem.CIMClass`):
            The class to be inspected for the qualifier defined by qname

          qname (:term:`string`):
            The qualifier for which we are searching

          scopes (tuple of booleans):
            A tuple containing a boolean value for each of the possible scopes
            (class, property, method, parameter)

        Returns:
          True if the qualifier with name quname is found in the elements where
          the scope is True. Otherwise, False is returned

        """
        # Test class scope
        if scopes[0] and qname in cls.qualifiers:
            return True

        # if property scope, test properties
        if scopes[1]:
            for prop in cls.properties.values():
                if qname in prop.qualifiers:
                    return True
        # If method scope, test methods and if parameter scope, test parameters
        if scopes[2]:
            for method in cls.methods.values():
                if qname in method.qualifiers:
                    return True
                if scopes[3]:
                    params = method.parameters
                    for param in params.values():
                        if qname in param.qualifiers:
                            return True
        return False

    # Test all classes in the input property for the defined filters.
    filtered_classes = []
    subclass_names = []
    # Build list of subclass names that will be used later as a filter on the
    # classes to be returned
    if 'subclass_of' in filters:
        try:
            subclass_names = get_subclass_names(
                classes,
                classname=filters['subclass_of'].optionvalue,
                deep_inheritance=True)
        except ValueError:
            raise click.ClickException(
                'Classname {} for "subclass-of" not found in returned classes.'
                .format(filters['subclass_of'].optionvalue))

    # Build a list of leaf class names that will be used later as a filter on
    # the classes to be returned.
    if 'leaf_classes' in filters:
        try:
            if subclass_names:
                clsx = [cls for cls in classes if cls.classname in
                        subclass_names]
                leafclass_names = get_leafclass_names(clsx)
            else:
                leafclass_names = get_leafclass_names(classes)

        except ValueError:
            raise click.ClickException(
                'Classname {} for "leaf_classes-of" not found in returned '
                'classes.'.format(filters['leaf_classes'].optionvalue))

    for cls in classes:
        show_class_list = []
        for filter_name, filter_ in filters.items():
            if filter_name == 'qualifier':
                option_value = filter_.optionvalue
                if class_has_qualifier(cls, filter_.qualifiername,
                                       filter_.scopes):
                    if filter_.qualifiername == 'version':
                        if filter_.qualifiername in cls.qualifiers:
                            cls_version = \
                                cls.qualifiers[filter_.qualifiername].value
                            version_val = parse_version_str(cls_version)
                            option_value = version_val >= filter_.optionvalue

                    show_class_list.append(option_value)
                else:
                    show_class_list.append(not option_value)

            elif filter_name == 'schema':
                show_class_list.append(
                    cls.classname.lower().startswith(filter_.optionvalue))
            elif filter_name == 'subclass_of':
                show_class_list.append(cls.classname in subclass_names)
            elif filter_name == 'leaf_classes':
                show_class_list.append(cls.classname in leafclass_names)

            else:
                assert False  # Future for other test_types

        # Show if all options are True for this class
        show_this_class = all(show_class_list)

        if show_this_class:
            # If returning instances, honor the names_only option
            if not names_only and not iq:
                cls.qualifiers = []
                for p in cls.properties.values():
                    p.qualifiers = []
                for m in cls.methods.values():
                    m.qualifiers = []
                    for p in m.parameters.values():
                        p.qualifiers = []
            filtered_classes.append(cls)

    # If names_only parameter create list of classnames
    if names_only:
        filtered_classes = [cls.classname for cls in filtered_classes]
    return filtered_classes


def enumerate_classes_filtered(context, classname, options):
    """
    Execute EnumerateClasses or EnumerateClassNames in a single namespace
    defined in options['namespace'] and return results.

    If any of the class qualifier filters are defined in the options parameter,
    enumerate the classes, filter the result for those parameters, and return
    only class names if --names-only set.

    This function may be executed by multiple command action functions with
    varying options in the options. Each option must be tested to validate
    that it exists in the options dictionary

    Parameters:

      context:  Click context

      classname:
        classname for the enumerate or None if all classes to be enumerated.

      options:Click options dictionary
        Options that form basis for this Enumerate and filter processing.

    Returns:
        List of classes or classnames that satisfy the criteria

    Raises:
        pywbem Error exceptions generated by EnumerateClassNames and
        enumerateClasses
    """
    namespace = options['namespace']
    conn = context.pywbem_server.conn
    filters = _build_filters_dict(conn, namespace, options)

    names_only = options.get('names_only', False)

    iq = options.get('no_qualifiers', True)

    # Force IncludeQualifier true if results are to be filtered since
    # the filter requires that qualifiers exist.
    request_iq = True if filters else iq

    local_only = options.get('local_only', False)
    deep_inheritance = options.get('deep_inheritance', True)
    include_classorigin = options.get('include_classorigin', True)

    if names_only and not filters:
        results = conn.EnumerateClassNames(
            ClassName=classname,
            namespace=namespace,
            DeepInheritance=deep_inheritance)
    else:
        results = conn.EnumerateClasses(
            ClassName=classname,
            namespace=namespace,
            LocalOnly=local_only,
            DeepInheritance=deep_inheritance,
            IncludeQualifiers=request_iq,
            IncludeClassOrigin=include_classorigin)
        if filters:
            results = _filter_classes(results, filters,
                                      names_only, iq)
    return results


def get_format_group(context, options):
    """
    Define the format groups allowed based on the options. This is particular
    to the class commands, largely because we do not have a table format
    for enumerate, get, associators, or references unless options such as
    summary, or names_only are set.
    """

    # Summary always output as TABLE
    if 'summary' in options and options['summary']:

        # This accounts for the fact that the results of a summary can be
        # either table or simply a string output
        if context.output_format and \
                output_format_is_table(context.output_format):
            return ['TABLE']

        # Temporary hack. We need another format group, i.e. txt or str
        # That displays in non-structured manner. Or drop this output
        # completely.
        return ['CIM']

    # Names_only may be output as Table or as CIM Object.
    if 'names_only' in options and options['names_only']:
        return ['CIM', 'TABLE']

    # otherwise only CIM allowed today.
    return ['CIM']


#####################################################################
#
#  Command functions for each of the commands in the class group
#
#####################################################################


def cmd_class_get(context, classname, options):
    """
    Get the class defined by the argument.

    Gets the class defined by CLASSNAME from the WBEM server and displays
    the class. If the class cannot be found, the server returns a CIMError
    exception.
    """
    conn = context.pywbem_server.conn
    format_group = get_format_group(context, options)
    output_format = validate_output_format(context.output_format, format_group)

    try:
        result_class = conn.GetClass(
            classname,
            namespace=options['namespace'],
            LocalOnly=options['local_only'],
            IncludeQualifiers=options['no_qualifiers'],
            IncludeClassOrigin=options['include_classorigin'],
            PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, result_class,
                            output_format=output_format)
    except Error as er:
        raise pywbem_error_exception(er)


def cmd_class_invokemethod(context, classname, methodname, options):
    """
    Create an instance and submit to a WBEM server
    """
    try:
        process_invokemethod(context, classname, methodname, options)
    except Error as er:
        raise pywbem_error_exception(er)


def cmd_class_enumerate(context, classname, options):
    """
        Enumerate the classes returning a list of classes from the WBEM server.
        That match the qualifier filter options
    """
    format_group = get_format_group(context, options)
    output_format = validate_output_format(context.output_format, format_group)

    try:
        results = enumerate_classes_filtered(context, classname, options)

        display_cim_objects(context, results, output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise pywbem_error_exception(er)


def cmd_class_references(context, classname, options):
    """
    Execute the references request operation to get references for
    the classname defined
    """
    conn = context.pywbem_server.conn
    if options['namespace']:
        classname = CIMClassName(classname, namespace=options['namespace'])

    format_group = get_format_group(context, options)
    output_format = validate_output_format(context.output_format, format_group)

    try:
        if options['names_only']:
            results = conn.ReferenceNames(
                classname,
                ResultClass=options['result_class'],
                Role=options['role'])
        else:
            results = conn.References(
                classname,
                ResultClass=options['result_class'],
                Role=options['role'],
                IncludeQualifiers=options['no_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise pywbem_error_exception(er)


def cmd_class_associators(context, classname, options):
    """
    Execute the references request operation to get references for
    the classname defined
    """
    conn = context.pywbem_server.conn

    if options['namespace']:
        classname = CIMClassName(classname, namespace=options['namespace'])

    format_group = get_format_group(context, options)
    output_format = validate_output_format(context.output_format, format_group)

    try:
        if options['names_only']:
            results = conn.AssociatorNames(
                classname,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'])
        else:
            results = conn.Associators(
                classname,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'],
                IncludeQualifiers=options['no_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise pywbem_error_exception(er)


def get_namespaces(context, namespaces):
    """
    Returns either the namespaces provided or if that is None, the set of
    namespaces that are defined in the wbem server as a list

    Raises:
        CIMError if status code not CIM_ERR_NOT_FOUND
    """
    conn = context.pywbem_server.conn
    wbem_server = context.pywbem_server.wbem_server

    ns_names = []

    # Return the provided namespace(s)
    if namespaces:
        return namespaces

    # Otherwise get all namespaces from server
    try:
        ns_names = wbem_server.namespaces
        ns_names.sort()
        return ns_names

    except ModelError:
        return [conn.default_namespace]
    except CIMError as ce:
        # allow processing to continue if no interop namespace
        if ce.status_code == CIM_ERR_NOT_FOUND:
            warning_msg('{}. Using default_namespace {}.'
                        .format(ce, conn.default_namespace))
            ns_names = [conn.default_namespace]
        return ns_names
    except Error as er:
        raise pywbem_error_exception(er)


def cmd_class_find(context, classname_glob, options):
    """
    Execute the command for enumerate classes, filter the results based on the
    option and display the result. The result is a list of classes/namespaces
    """

    output_format = validate_output_format(context.output_format, 'TABLE')

    context.spinner_stop()
    namespaces = get_namespaces(context, options['namespace'])

    try:
        # Define sort by namespace name if --sort, or  othersise by count
        sort_col = 0 if options['sort'] else 1

        # Dictionary key is namespaces, value is list of classes
        names_dict = NocaseDict()
        if namespaces:
            for ns in namespaces:
                # Uses di True, and names_only. Build options for enumerate
                options['deep_inheritance'] = True
                options['namespace'] = ns
                options['names_only'] = True

                # enumerate filtered for classnames and add to namespace in
                # dictionary
                classnames = enumerate_classes_filtered(context, None, options)
                names_dict[ns] = filter_namelist(classname_glob, classnames)

        context.spinner_stop()

        # If summary set, generate output of only count of classes for
        # each namespace
        if 'summary' in options and options['summary']:
            rows = [[ns, len(names_dict[ns])] for ns in names_dict]
            rows.sort(key=lambda x: x[sort_col])

            if output_format_is_table(context.output_format):
                headers = ['Namespace', 'Class count']
                click.echo(format_table(rows,
                                        headers,
                                        title='Find class counts {}'.
                                        format(classname_glob),
                                        table_format=output_format))
            else:
                for ns, count in rows:
                    click.echo('  {}: {}'. format(ns, count))
            return

        # Not summary. Build rows of namespace, classname for each namespace
        # and sort by names if --sort set
        rows = []
        for ns_name in names_dict:
            ns_rows = [[ns_name, name] for name in names_dict[ns_name]]
            # sort by classname if sort option defined, else by namespace
            ns_rows.sort(key=lambda x: x[sort_col])
            rows.extend(ns_rows)

        context.spinner_stop()

        if output_format_is_table(context.output_format):
            headers = ['Namespace', 'Classname']
            click.echo(
                format_table(rows, headers,
                             title='Find class {}'.format(classname_glob),
                             table_format=output_format))
        else:
            # Display function to display classnames returned with
            # their namespaces in the form <namespace>:<classname>
            for row in rows:
                click.echo('  {}: {}'.format(row[0], row[1]))

    except Error as er:
        raise pywbem_error_exception(er)


def get_class_hierarchy(conn, classname, namespace, superclasses=None):
    """
    Get the class hierarchy from the server, either the superclasses
    associated or subclasses with classname .  If getting subclasses
    the classname parameter may be None which requests that the complete
    class hiearchy be retrieved.

      Parameters:
        conn (:class:`~pywbem.WBEMConnection` or subclass):
            Current connection to a WBEM server.

        classname(:term:`string):
            classname if the tree is to be initiated from
            within the class hiearchy. May be None

        namespace(:term:`string):
          Namespace to use to acquire classes from WBEM server

        superclasses (:class:`py:bool`):
            If `True` display the superclasses of classname. If not True
            display the subclasses. The default is None (display subclasses).

      Returns:
        tuple of  classes, classnames where
          classes (list of :class:`~pywbem.CIMClass`): that are either the
          superclasses or the subclasses of classname.

      Raises:
        CIM_Error:
    """

    try:
        if superclasses:
            # classname must exist for superclass tree
            assert classname is not None

            # get the superclasses into a list
            klass = conn.GetClass(classname, namespace=namespace)
            classes = []
            classes.append(klass)
            while klass.superclass:
                klass = conn.GetClass(klass.superclass, namespace=namespace)
                classes.append(klass)
        else:
            # get the subclass hierarchy either complete or starting at the
            # optional CLASSNAME. Gets minimum data from server to define
            # class, superclass data
            classes = conn.EnumerateClasses(ClassName=classname,
                                            namespace=namespace,
                                            LocalOnly=True,
                                            IncludeQualifiers=True,
                                            DeepInheritance=True)

    except Error as er:
        raise pywbem_error_exception(er)

    return classes


def cmd_class_tree(context, classname, options):
    """
    Execute the command to display graphical tree of either superclasses or
    subclasses of classname. If classname is None, display tree starting from
    the class hierarchy root.
    Tree is displayed on the console as a left-justified tree using the
    asciitree library.
    The --superclasses option determines if the superclass tree or the
    subclass tree is displayed.
    """
    conn = context.pywbem_server.conn

    superclasses = options['superclasses']

    # Classname must exist as starting point for superclass tree.
    if superclasses and classname is None:
        raise click.ClickException('CLASSNAME argument required for '
                                   '--superclasses option')

    classes = get_class_hierarchy(conn, classname, options['namespace'],
                                  superclasses)

    # If showing superclasses, set classname to None for the display
    # to add the 'root' as top class.
    if superclasses:
        classname = None

    # Display the list of classes as a tree. The classname is the top
    # of the tree.
    context.spinner_stop()
    display_class_tree(classes, classname, show_detail=options['detail'])


def cmd_class_delete(context, classname, options):
    """Delete a class from the WBEM server repository"""
    conn = context.pywbem_server.conn

    include_instances = options['include_instances']
    if options['force']:
        include_instances = options['force']
        pywbemcliwarn("The --force / -f option has been deprecated and "
                      "will be removed in a future version. Use "
                      "--include-instances instead.", DeprecationWarning)
    dry_run = options['dry_run']
    dry_run_prefix = "Dry run: " if dry_run else ""

    namespace = options['namespace'] or conn.default_namespace

    try:
        instnames = conn.EnumerateInstanceNames(
            ClassName=classname, namespace=namespace)
    except Error as exc:
        raise pywbem_error_exception(
            exc, "Cannot enumerate instance names of class {} in "
            "namespace {}".format(classname, namespace))

    if instnames and not include_instances:
        raise click.ClickException(
            "Cannot delete class {} because it has {} instances".
            format(classname, len(instnames)))

    depending_cln_list = depending_classnames(
        classname, namespace, conn)

    if depending_cln_list:
        raise click.ClickException(
            "Cannot delete class {} because these classes depend on it: {}".
            format(classname, ', '.join(depending_cln_list)))

    context.spinner_stop()
    if include_instances:
        for instname in instnames:
            if not dry_run:
                try:
                    conn.DeleteInstance(instname)
                except Error as exc:
                    raise pywbem_error_exception(
                        exc, "Cannot delete instance {}".format(instname))
            click.echo('{}Deleted instance {}'.format(dry_run_prefix, instname))

    if not dry_run:
        try:
            conn.DeleteClass(classname)
        except Error as exc:
            raise pywbem_error_exception(
                exc, "Cannot delete class {} in namespace {}".
                format(classname, namespace))
    click.echo('{}Deleted class {}'.format(dry_run_prefix, classname))



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click command definition for the server command group which includes
cmds for inspection and management of the objects defined by the pywbem
server class including namespaces, WBEMServer information, and profile
information.

NOTE: Commands are ordered in help display by their order in this file.
"""

from __future__ import absolute_import, print_function

import os
import sys
import click

from pywbem import Error, MOFCompiler
from pywbem._mof_compiler import MOFWBEMConnection, MOFCompileError

from .pywbemcli import cli
from ._common import format_table, pywbem_error_exception, \
    validate_output_format, display_text, \
    CMD_OPTS_TXT, GENERAL_OPTS_TXT, SUBCMD_HELP_TXT
from ._common_options import add_options, help_option, namespace_option
from ._click_extensions import PywbemcliGroup, PywbemcliCommand
from ._cmd_namespace import cmd_namespace_list, cmd_namespace_interop
from ._utils import pywbemcliwarn

# NOTE: A number of the options use double-dash as the short form.  In those
# cases, a third definition of the options without the double-dash defines
# the corresponding option name, ex. 'include_qualifiers'. It should be
# defined with underscore and not dash

# Issue 224 - Exception in prompt-toolkit with python 2.7. Caused because
# with prompt-toolkit 2 + the completer requires unicode and click_repl not
# passing help as unicode in options as unicode
# NOTE: Insure that all option help attributes are unicode to get around this
#       issue

#
#   Common option definitions for server group
#

mof_include_option = [              # pylint: disable=invalid-name
    click.option('--include', '-I', metavar='INCLUDEDIR', multiple=True,
                 help=u'Path name of a MOF include directory. '
                 'May be specified multiple times.')]

mof_dry_run_option = [              # pylint: disable=invalid-name
    click.option('--dry-run', '-d', is_flag=True, default=False,
                 help=u'Enable dry-run mode: Don\'t actually modify the '
                 'server. Connection to the server is still required for '
                 'reading.')]


@cli.group('server', cls=PywbemcliGroup, options_metavar=GENERAL_OPTS_TXT,
           subcommand_metavar=SUBCMD_HELP_TXT)
@add_options(help_option)
def server_group():
    """
    Command group for WBEM servers.

    This command group defines commands to inspect and manage core components
    of a WBEM server including server attributes, namespaces, compiling MOF,
    the Interop namespace, management profiles, and access to profile central
    instances.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'server' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@server_group.command('namespaces', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def server_namespaces(context):
    """
    List the namespaces of the server (deprecated).

    The Interop namespace must exist on the server.

    Deprecated: The 'server namespaces' command is deprecated and will be
    removed in a future version. Use the 'namespace list' command instead.
    """
    pywbemcliwarn(
        "The 'server namespaces' command is deprecated and will be removed in "
        "a future version. Use the 'namespace list' command instead.",
        DeprecationWarning)
    context.execute_cmd(lambda: cmd_namespace_list(context))


@server_group.command('interop', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def server_interop(context):
    """
    Get the Interop namespace of the server (deprecated).

    The Interop namespace must exist on the server.

    Deprecated: The 'server interop' command is deprecated and will be removed
    in a future version. Use the 'namespace interop' command instead.
    """
    pywbemcliwarn(
        "The 'server interop' command is deprecated and will be removed in "
        "a future version. Use the 'namespace interop' command instead.",
        DeprecationWarning)
    context.execute_cmd(lambda: cmd_namespace_interop(context))


@server_group.command('brand', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def server_brand(context):
    """
    Get the brand of the server.

    Brand information is defined by the server implementor and may or may
    not be available. Pywbem attempts to collect the brand information from
    multiple sources.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_server_brand(context))


@server_group.command('info', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def server_info(context):
    """
    Get information about the server.

    The information includes CIM namespaces and server brand.
    """
    context.execute_cmd(lambda: cmd_server_info(context))


@server_group.command('add-mof', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@click.argument('moffiles', metavar='MOFFILE', type=click.Path(),
                nargs=-1, required=True)
@add_options(namespace_option)
@add_options(mof_include_option)
@add_options(mof_dry_run_option)
@add_options(help_option)
@click.pass_obj
def server_add_mof(context, **options):
    """
    Compile MOF and add/update CIM objects in the server.

    The MOF files are specified with the MOFFILE argument, which may be
    specified multiple times. The minus sign ('-') specifies the standard
    input.

    Initially, the target namespace is the namespace specified with the
    --namespace option or if not specified the default namespace of the
    connection. If the MOF contains '#pragma namespace' directives, the target
    namespace will be changed accordingly.

    MOF include files (specified with the '#pragma include' directive) are
    searched first in the directory of the including MOF file, and then in
    the directories specified with the --include option.

    Any CIM objects (instances, classes and qualifiers) specified in the MOF
    files are created in the server, or modified if they already exist in the
    server.

    The global --verbose option will show the CIM objects that are created or
    modified.
    """
    context.execute_cmd(lambda: cmd_server_add_mof(context, options))


@server_group.command('remove-mof', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@click.argument('moffiles', metavar='MOFFILE', type=click.Path(),
                nargs=-1, required=True)
@add_options(namespace_option)
@add_options(mof_include_option)
@add_options(mof_dry_run_option)
@add_options(help_option)
@click.pass_obj
def server_remove_mof(context, **options):
    """
    Compile MOF and remove CIM objects from the server.

    The MOF files are specified with the MOFFILE argument, which may be
    specified multiple times. The minus sign ('-') specifies the standard
    input.

    Initially, the target namespace is the namespace specified with the
    --namespace option or if not specified the default namespace of the
    connection. If the MOF contains '#pragma namespace' directives, the target
    namespace will be changed accordingly.

    MOF include files (specified with the '#pragma include' directive) are
    searched first in the directory of the including MOF file, and then in
    the directories specified with the --include option.

    Any CIM objects (instances, classes and qualifiers) specified in the MOF
    files are deleted from the server.

    The global --verbose option will show the CIM objects that are removed.
    """
    context.execute_cmd(lambda: cmd_server_remove_mof(context, options))


###############################################################
#         Server cmds
###############################################################


def cmd_server_brand(context):
    """
    Display product and version info of the current WBEM server
    """
    wbem_server = context.pywbem_server.wbem_server
    output_format = validate_output_format(context.output_format, 'TEXT')

    try:
        brand = wbem_server.brand
        context.spinner_stop()

        display_text(brand, output_format)

    except Error as er:
        raise pywbem_error_exception(er)


def cmd_server_info(context):
    """
    Display general overview of info from current WBEM server
    """
    wbem_server = context.pywbem_server.wbem_server
    output_format = validate_output_format(context.output_format, 'TABLE')

    try:
        # Execute the namespaces to force contact with server before
        # turning off the spinner.
        namespaces = sorted(wbem_server.namespaces)
        context.spinner_stop()

        rows = []
        headers = ['Brand', 'Version', 'Interop Namespace', 'Namespaces']
        sep = '\n' if namespaces and len(namespaces) > 3 else ', '
        namespaces = sep.join(namespaces)

        rows.append([wbem_server.brand, wbem_server.version,
                     wbem_server.interop_ns,
                     namespaces])
        click.echo(format_table(rows, headers,
                                title='Server General Information',
                                table_format=output_format))

    except Error as er:
        raise pywbem_error_exception(er)


def cmd_server_add_mof(context, options):
    """
    Compile MOF and add/update CIM objects in the server.
    """
    conn = context.pywbem_server.conn

    try:

        context.spinner_stop()

        # Define the connection to be used by the MOF compiler.
        # MOFWBEMConnection writes resulting CIM objects to a local store
        # but reads from the connection.
        if options['dry_run']:
            comp_handle = MOFWBEMConnection(conn=conn)
        else:
            comp_handle = conn

        if options['dry_run']:
            print('Executing in dry-run mode')

        include_dirs = []
        for idir in options['include']:
            if not os.path.isabs(idir):
                idir = os.path.abspath(idir)
            include_dirs.append(idir)
        for moffile in options['moffiles']:
            if moffile != '-':
                mofdir = os.path.dirname(moffile)
                if not os.path.isabs(mofdir):
                    mofdir = os.path.abspath(mofdir)
                for idir in include_dirs:
                    if mofdir.startswith(idir):
                        break
                else:
                    include_dirs.append(mofdir)

        mofcomp = MOFCompiler(handle=comp_handle, search_paths=include_dirs,
                              verbose=context.verbose)

        for moffile in options['moffiles']:
            if moffile == '-':
                mofstr = sys.stdin.read()  # bytes in py2 / text in py3
                if context.verbose:
                    print('Compiling MOF from standard input')
                # The defaulting to the connection default namespace is handled
                # inside of the MOF compiler.
                mofcomp.compile_string(mofstr, options['namespace'])
            else:
                if not os.path.isabs(moffile):
                    moffile = os.path.abspath(moffile)
                if context.verbose:
                    print('Compiling MOF file {0}'.format(moffile))
                # The defaulting to the connection default namespace is handled
                # inside of the MOF compiler.
                mofcomp.compile_file(moffile, options['namespace'])

    # If MOFCompileError, exception already logged by compile_string().
    except MOFCompileError:
        raise click.ClickException("Compile failed.")

    # Otherwise display the exception itself
    except Error as exc:
        raise pywbem_error_exception(exc)


def cmd_server_remove_mof(context, options):
    """
    Compile MOF and remove CIM objects from the server.
    """
    conn = context.pywbem_server.conn

    try:

        context.spinner_stop()

        # Define the connection to be used by the MOF compiler.
        # MOFWBEMConnection writes resulting CIM objects to a local store
        # but reads from the connection.
        comp_handle = MOFWBEMConnection(conn=conn)

        if options['dry_run']:
            print('Executing in dry-run mode')

        include_dirs = []
        for idir in options['include']:
            if not os.path.isabs(idir):
                idir = os.path.abspath(idir)
            include_dirs.append(idir)
        for moffile in options['moffiles']:
            if moffile != '-':
                mofdir = os.path.dirname(moffile)
                if not os.path.isabs(mofdir):
                    mofdir = os.path.abspath(mofdir)
                for idir in include_dirs:
                    if mofdir.startswith(idir):
                        break
                else:
                    include_dirs.append(mofdir)

        # verbose messages are displayed by rollback()
        mofcomp = MOFCompiler(handle=comp_handle, search_paths=include_dirs,
                              verbose=False)

        for moffile in options['moffiles']:
            if moffile == '-':
                mofstr = sys.stdin.read()  # bytes in py2 / text in py3
                if context.verbose:
                    print('Compiling MOF from standard input into cache')
                # The defaulting to the connection default namespace is handled
                # inside of the MOF compiler.
                mofcomp.compile_string(mofstr, options['namespace'])
            else:
                if not os.path.isabs(moffile):
                    moffile = os.path.abspath(moffile)
                if context.verbose:
                    print('Compiling MOF file {0} into cache'.format(moffile))
                # The defaulting to the connection default namespace is handled
                # inside of the MOF compiler.
                mofcomp.compile_file(moffile, options['namespace'])

        # rollback the compiled objects to remove them from the target.
        if not options['dry_run']:
            if context.verbose:
                print('Deleting CIM objects found in MOF...')
            comp_handle.rollback(verbose=context.verbose)
        else:
            if context.verbose:
                print('No deletions will be shown in dry-run mode')
    # If MOFCompileError, exception already logged by compile_string().
    except MOFCompileError:
        raise click.ClickException("Compile failed.")
    except Error as exc:
        raise pywbem_error_exception(exc)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
pybemcli context object. This is the common object for click command calls for
pywbemcli context information.

It contains data that is set in the top level and used in command calls

This object is attached to the Click Context in pywbemcli.py
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import click
import click_spinner

from ._common import format_table, validate_output_format


class ContextObj(object):
    # pylint: disable=useless-object-inheritance, too-many-instance-attributes
    """
        Manage the pywbemcli context that is carried within the Click
        context object in the obj parameter. This is the object that
        communicates between the cli commands and command groups. It contains
        the information that is common to the multiple click commands
    """

    spinner_envvar = 'PYWBEMCLI_SPINNER'

    # pylint: disable=unused-argument
    def __init__(self, pywbem_server, output_format, use_pull,
                 pull_max_cnt, timestats, log, verbose, pdb,
                 warn, connections_repo, interactive_mode,
                 close_interactive_server):
        """
        Parameters:

          pywbem_server :class:`PywbemServer`):
            PywbemServer instance to be used for connection

          output_format (:term:`string or None):
            String representing the type of output from the --output-format
            general option or None if the default is to be used.

          use_pull (:class:`py:bool` or None):
            If boolean defines whether pull operations are to be used.  If
            None, the pywbem client will decide.           `

          pull_max_cnt(:term:`integer`):
            Positive integer represenging the max size of each pull
            operation response

          timestats (:class:`py:bool`):
            See timestats general option

          log (:term:`string or None):
            String defining the characteristics of log output for each
            WBEM operation or None if logging is disabled.

          verbose (:class:`py:bool`):
            See verbose general option

          pdb (:class:`py:bool`):
            See pdb general option

          warn (:class:`py:bool`):
            See warn general option

          connections_repo:

          interactive_mode (:class:`py:bool`):
            If True, pywbemcli is in interactive mode.

          close_interactive_server (:class:`py:bool`):
            Flag that defines interactive command with a server definition
            that must be disconnected after the command

        """

        self._pywbem_server = pywbem_server
        self._output_format = output_format
        self._use_pull = use_pull
        self._pull_max_cnt = pull_max_cnt
        self.timestats = timestats   # has setter method
        self._log = log
        self._verbose = verbose
        self._pdb = pdb
        # warn included only to insure that original cmd line input flag
        # maintained through interactive session.
        self._warn = warn
        self._connections_repo = connections_repo
        self.interactive_mode = interactive_mode
        self._close_interactive_server = close_interactive_server

        self._spinner_enabled = None  # Deferred init in getter
        self._spinner_obj = click_spinner.Spinner()

    def __repr__(self):
        return 'ContextObj(at {:08x}, pywbem_server={!r}, outputformat={}, ' \
               'use_pull={}, pull_max_cnt={}, timestats={}, verbose={}, ' \
               'spinner_enabled={}, interactive_mode={}' \
               .format(id(self), self._pywbem_server, self.output_format,
                       self.use_pull, self.pull_max_cnt, self.timestats,
                       self.verbose, self.spinner_enabled,
                       self._interactive_mode)

    @property
    def output_format(self):
        """
        :term:`string`: String defining the output format requested.  This may
        be None meaning that the default format should be used or may be
        one of the values in the TABLE_FORMATS variable.
        """
        return self._output_format

    @property
    def timestats(self):
        """
        :term:`string`: Output format to be used.
        """
        return self._timestats

    @timestats.setter
    def timestats(self, value):
        """Setter method; for a description see the getter method."""
        # pylint: disable=attribute-defined-outside-init
        self._timestats = value

    @property
    def interactive_mode(self):
        """
        :class:`py:bool`: 'True' if in interactive mode and 'False' if
        in the command mode.
        """
        return self._interactive_mode

    @interactive_mode.setter
    def interactive_mode(self, mode):
        """Setter method; for a description see the getter method."""
        assert isinstance(mode, bool)
        # pylint: disable=attribute-defined-outside-init
        self._interactive_mode = mode

    @property
    def use_pull(self):
        """
        :term:`string`: Choice of whether pull, traditional or either type
        of operation is to be used for the instance enumerates, references,
        or associator commands.
        """
        return self._use_pull

    @property
    def pull_max_cnt(self):
        """
        :term:`string`: Maximum number of objects to be returne for pull op.
        """
        return self._pull_max_cnt

    @property
    def log(self):
        """
        :term:`string`: log definition from cmd line
        """
        return self._log

    @property
    def pdb(self):
        """
        bool: Indicates whether to break in the debugger.
        """
        return self._pdb

    @property
    def warn(self):
        """
        bool: Indicates whether to enable Python warnings.
        """
        return self._warn

    @property
    def connections_repo(self):
        """
        :class:`ConnectionRepository` instance defining at least the filename
        of the Connections repository. The connections_repo may not be
        loaded at this point.
        """
        return self._connections_repo

    def is_connected(self):
        """
        bool: Indicates whether currently connected to a WBEM server. Where
        connected implies both the existence of the PywbemServer instance
        and that the connection to a server has been made.

        That is the case as soon as commands have been executed that
        communicate with the server.
        """
        return self._pywbem_server and self._pywbem_server.connected

    @property
    def pywbem_server(self):
        """
        :class:`PywbemServer`: Current PywbemServer object for this context.

        Return the PywbemServer object if it already exists.

        If the pywbem_server attribute does not exist, generate a click
        exception. If no server is specified, `None` is returned.

        This attribute is settable.
        """
        if self._pywbem_server:
            return self._pywbem_server

        ctx = click.get_current_context()
        if ctx.parent:
            ctx = ctx.parent
        cmd = "{} {}".format(ctx.info_name or "",
                             ctx.invoked_subcommand or "")
        raise click.ClickException(
            'No  current server for command "{}" that requires a WBEM server. '
            'Specify a server with the "--server", "--mock-server", or"--name" '
            'general option, by setting the corresponding environment '
            'variables, or in interactive mode '
            'use "connection select" to define a target server'.
            format(cmd))

    @pywbem_server.setter
    def pywbem_server(self, value):
        """Setter method; for a description see the getter method."""
        # pylint: disable=attribute-defined-outside-init
        self._pywbem_server = value

    @property
    def spinner_enabled(self):
        """
        :class:`py:bool`: Indicates and controls whether the spinner is enabled.

        If the spinner is enabled, subcommands will display a spinning wheel
        while waiting for completion.

        This attribute can be modified.

        The initial state of the spinner is enabled, but it can be disabled by
        setting the {0} environment variable to 'false', '0', or the empty
        value.
        """.format(self.spinner_envvar)

        # Deferred initialization
        if self._spinner_enabled is None:
            value = os.environ.get(self.spinner_envvar, None)
            if value is None:
                # Default if not set
                self._spinner_enabled = True
            elif value == '0' or value == '' or value.lower() == 'false':
                self._spinner_enabled = False
            else:
                self._spinner_enabled = True

        return self._spinner_enabled

    @spinner_enabled.setter
    def spinner_enabled(self, enabled):
        """Setter method; for a description see the getter method."""
        # pylint: disable=attribute-defined-outside-init
        self._spinner_enabled = enabled

    def spinner_start(self):
        """
        Start the spinner, if the spinner is enabled.
        """
        if self.spinner_enabled:
            self._spinner_obj.start()

    def spinner_stop(self):
        """
        Stop the spinner, if the spinner is enabled.
        """
        if self.spinner_enabled:
            self._spinner_obj.stop()

    @property
    def verbose(self):
        """
        :bool:` '~click.verbose.
        """
        return self._verbose

    def pywbem_server_exists(self):
        """
        Return True if a pywbem_server is defined.  This method allows testing
        for the existence of a PywbemServer object without causing an exception.
        The normal method is to simply access the property. In that case,
        an exception will be generated if there is no PywbemServer defined.

        Returns:
            :class": `boolean True if an instance of PywbemServer is defined.
        """
        return self._pywbem_server is not None

    def execute_cmd(self, cmd):
        """
        Call the cmd executor defined by cmd with the spinner. If the
        WBEM server object has not been created it is created so that this WBEM
        server can be used for interactive commands.

        This method is called by every command execution to setup and
        execute the command. Thus, each command definition MUST have the line
        similar to:

        context.execute_cmd(lambda: cmd_instance_query(context, query, options))
        """
        # Env.var PYWBEMCLI_DIAGNOSTICS turns on diagnostic prints for developer
        # use and is therefore not documented.
        if os.getenv('PYWBEMCLI_DIAGNOSTICS'):
            ctx = click.get_current_context()
            click.echo('DIAGNOSTICS-CMD: info_name={!r}, subcommand={!r}, '
                       'command={!r}, params={!r}'.
                       format(ctx.info_name, ctx.invoked_subcommand,
                              ctx.command,
                              ctx.params))
            display_click_context_parents(display_attrs=True)

        if not self.pdb:
            self.spinner_start()
        try:
            if self.pdb:
                import pdb  # pylint: disable=import-outside-toplevel
                pdb.set_trace()  # pylint: disable=no-member

            cmd()  # The pywbemcli command function call.

        finally:
            if not self.pdb:
                self.spinner_stop()

            # Issue statistics if requested and if the command used a conn.
            if self.timestats and self.is_connected():
                context = click.get_current_context()
                click.echo(self.format_statistics(
                    self.pywbem_server.conn.statistics, context.obj))

            # Close any existing connection if in command mode or if the
            # close_interactive_server flag is set
            if not self.interactive_mode or self._close_interactive_server:
                if self.is_connected():
                    self.pywbem_server.disconnect()

    def format_statistics(self, statistics, context):
        # pylint: disable=no-self-use
        """
        Table formatted output of client statistics
        """
        output_fmt = validate_output_format(context.output_format, 'TABLE')

        snapshot = sorted(statistics.snapshot(),
                          key=lambda item: item[1].avg_time,
                          reverse=True)

        header = ['Operation', 'Count', 'Errors',
                  'Client Time\n[ms]',
                  'Server Time\n[ms]',
                  'Request Size\n[B]',
                  'Response Size\n[B]']

        rows = []
        for name, stats in snapshot:
            row = [name, stats.count, stats.exception_count,
                   '{:.3f}'.format(stats.avg_time * 1000),
                   '{:.3f}'.format(stats.avg_server_time * 1000),
                   '{:.0f}'.format(stats.avg_request_len),
                   '{:.0f}'.format(stats.avg_reply_len)]
            rows.append(row)

        click.echo(format_table(
            rows, header, title='Client statistics',
            table_format=output_fmt))

    @staticmethod
    def update_root_click_context(ctx_obj):
        """
        Method to update the click root context with a new context
        object.  This makes the values in this context universal for all
        future commands within an interactive session.
        This is a static method and gets the current click context and
        root context from click itself.
        """
        ctx = click.get_current_context()
        root = ctx.find_root()
        root.obj = ctx_obj


#
#   Debug tools to help understanding the click context.  These are only
#   to help analyzing the click_context (ctx) and primarily in the
#   interactive mode when several levels of click context can exist
#
def display_click_context(ctx, msg=None, display_attrs=True):
    """
    Debug function displays attributes of click context
    This is a diagnostic for developer use
    """

    attrs = vars(ctx)
    if not msg:
        msg = "CLICK_CONTEXT"
    if not display_attrs:
        click.echo(ctx.obj)
    else:
        click.echo('{0} {1}, attrs:\n    {2}'.format(
            msg, ctx,
            '\n    '.join(
                '{0}: {1}'.format(i, v)
                for (i, v) in sorted(attrs.items()))))


def display_click_context_parents(display_attrs=False):
    """
    Display the current click context and its all of its parents.
    This is a diagnostic for developer use
    """
    ctx = click.get_current_context()
    disp_ctx = ctx
    # display in context top-down order
    stack = []
    # Create list of ctx parents
    while disp_ctx is not None:
        stack.append(disp_ctx)
        disp_ctx = disp_ctx.parent

    # pop off of stack to get reverse order
    level = 0
    while stack:
        disp_ctx = stack.pop()
        display_click_context(
            disp_ctx, msg='DIAGNOSTICS-CTX: Context at parent level {}:'.
            format(level), display_attrs=display_attrs)
        level += 1



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click Command definition for the qualifer command group which includes
cmds for get and enumerate for CIM qualifier types.

NOTE: Commands are ordered in help display by their order in this file.
"""

from __future__ import absolute_import, print_function

import click

from pywbem import Error

from .pywbemcli import cli
from ._common import sort_cimobjects, \
    pywbem_error_exception, validate_output_format, \
    CMD_OPTS_TXT, GENERAL_OPTS_TXT, SUBCMD_HELP_TXT
from ._display_cimobjects import display_cim_objects
from ._common_options import add_options, namespace_option, summary_option, \
    help_option
from ._click_extensions import PywbemcliGroup, PywbemcliCommand

# Issue 224 - Exception in prompt-toolkit with python 2.7. Caused because
# with prompt-toolkit 2 + the completer requires unicode and click_repl not
# passing help as unicode in options as unicode
# NOTE: Insure that all option help attributes are unicode to get around this
#       issue


@cli.group('qualifier', cls=PywbemcliGroup, options_metavar=GENERAL_OPTS_TXT,
           subcommand_metavar=SUBCMD_HELP_TXT)
@add_options(help_option)
def qualifier_group():
    """
    Command group for CIM qualifier declarations.

    This command group defines commands to inspect and delete CIM qualifier
    declarations in the WBEM Server.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'qualifier' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@qualifier_group.command('get', cls=PywbemcliCommand,
                         options_metavar=CMD_OPTS_TXT)
@click.argument('qualifiername', type=str, metavar='QUALIFIERNAME',
                required=True,)
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def qualifier_get(context, qualifiername, **options):
    """
    Get a qualifier declaration.

    Get a CIM qualifier declaration (QUALIFIERNAME argument) in a CIM
    namespace (--namespace option). If no namespace was specified, the default
    namespace of the connection is used.

    In the output, the qualifier declaration will formatted as defined by the
    --output-format general option.
    """
    context.execute_cmd(lambda: cmd_qualifier_get(context, qualifiername,
                                                  options))


@qualifier_group.command('delete', cls=PywbemcliCommand,
                         options_metavar=CMD_OPTS_TXT)
@click.argument('qualifiername', type=str, metavar='QUALIFIERNAME',
                required=True,)
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def qualifier_delete(context, qualifiername, **options):
    """
    Delete a qualifier declaration.

    Delete a CIM qualifier declaration (QUALIFIERNAME argument) in a CIM
    namespace (--namespace option). If no namespace was specified, the default
    namespace of the connection is used.

    This command executes the DeleteQualifier operation against the WBEM server
    and leaves it to the WBEM server to reject the operation if any classes
    in the namespace use the qualifier.

    In the output, the qualifier declaration will formatted as defined by the
    --output-format general option.
    """
    context.execute_cmd(lambda: cmd_qualifier_delete(
        context, qualifiername, options))


@qualifier_group.command('enumerate', cls=PywbemcliCommand,
                         options_metavar=CMD_OPTS_TXT)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(help_option)
@click.pass_obj
def qualifier_enumerate(context, **options):
    """
    List the qualifier declarations in a namespace.

    Enumerate the CIM qualifier declarations in the specified CIM namespace
    (--namespace option). If no namespace was specified, the default namespace
    of the connection is used.

    In the output, the qualifier declaration will formatted as defined by the
    --output-format general option.
    """
    context.execute_cmd(lambda: cmd_qualifier_enumerate(context, options))


####################################################################
#   Qualifier declaration command processing functions
#####################################################################


def cmd_qualifier_get(context, qualifiername, options):
    """
    Execute the command for get qualifier and display result
    """
    conn = context.pywbem_server.conn
    output_format = validate_output_format(context.output_format, ['CIM',
                                                                   'TABLE'])

    try:
        qual_decl = conn.GetQualifier(qualifiername,
                                      namespace=options['namespace'])

        display_cim_objects(context, qual_decl, output_format)

    except Error as er:
        raise pywbem_error_exception(er)


def cmd_qualifier_delete(context, qualifiername, options):
    """
    Execute the command for delete qualifier and display result
    """
    conn = context.pywbem_server.conn
    try:
        conn.DeleteQualifier(qualifiername, namespace=options['namespace'])
        if context.verbose:
            context.spinner_stop()
            click.echo('Deleted qualifier type {}.'.format(qualifiername))
    except Error as er:
        raise pywbem_error_exception(er)


def cmd_qualifier_enumerate(context, options):
    """
    Execute the command for enumerate qualifiers and desplay the result.
    """
    conn = context.pywbem_server.conn
    output_format = validate_output_format(context.output_format, ['CIM',
                                                                   'TABLE'])

    try:
        qual_decls = sort_cimobjects(conn.EnumerateQualifiers(
            namespace=options['namespace']))

        display_cim_objects(context, qual_decls, output_format,
                            summary=options['summary'])

    except Error as er:
        raise pywbem_error_exception(er)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click Command definition for the connection command group which includes
cmds to list add, delete, show, and test server definitions in the
connection information.  This also maintains a definition of servers on
disk which is kept in sync with changes made with the add/delete commands.

NOTE: Commands are ordered in help display by their order in this file.
"""

from __future__ import absolute_import, print_function

from copy import deepcopy
import click
import six

from pywbem import Error, CIMError, CIM_ERR_NOT_SUPPORTED

from .pywbemcli import cli
from ._common import CMD_OPTS_TXT, GENERAL_OPTS_TXT, \
    SUBCMD_HELP_TXT, pick_one_from_list, format_table, \
    pywbem_error_exception, validate_output_format, fold_strings, \
    output_format_is_table
from ._connection_repository import ConnectionsFileError
from ._common_options import add_options, help_option
from ._pywbem_server import PywbemServer
from ._context_obj import ContextObj
from ._click_extensions import PywbemcliGroup, PywbemcliCommand

# Issue 224 - Exception in prompt-toolkit with python 2.7. Caused because
# with prompt-toolkit 2 + the completer requires unicode and click_repl not
# passing help as unicode in options as unicode
# NOTE: Insure that all option help attributes are unicode to get around this
#       issue


@cli.group('connection', cls=PywbemcliGroup, options_metavar=GENERAL_OPTS_TXT,
           subcommand_metavar=SUBCMD_HELP_TXT)
@add_options(help_option)
def connection_group():
    """
    Command group for WBEM connection definitions.

    This command group defines commands to manage persistent WBEM connection
    definitions that have a name. The names of these connection definitions
    can then be used as shorthand for the WBEM server or mock server via the
    '--name' general option.

    The connection definitions are stored in a connections file. By default,
    the connections file is '.pywbemcli_connections.yaml' in the
    user's home directory. The location of the user's home directory depends on
    the operating system used. It is determined with Python's
    'os.path.expanduser("~")', which works on all operating systems including
    Windows. The default path name of the connections file can be overwritten
    using the 'PYWBEMCLI_CONNECTIONS_FILE' environment variable, or with the
    '--connections-file' general option.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'connection' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@connection_group.command('export', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def connection_export(context):
    """
    Export the current connection.

    Display commands that set pywbemcli environment variables to the parameters
    of the current connection.

    Examples:

      pywbemcli --name srv1 connection export

      pywbemcli --server https://srv1 --user me --password pw connection export
    """
    context.execute_cmd(lambda: cmd_connection_export(context))


@connection_group.command('show', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@click.argument('name', type=str, metavar='NAME', required=False)
@click.option('--show-password', is_flag=True,
              default=False,
              help=u'If set, show existing password in results. Otherwise, '
                   u'password is masked')
@add_options(help_option)
@click.pass_obj
def connection_show(context, name, **options):
    """
    Show a WBEM connection definition or the current connection.

    Show the name and attributes of a WBEM connection definition or the
    current connection, as follows:

    * If the NAME argument is specified, display the connection information
      with that name from the connections file or the current connection if it
      is the same name.

    * If the NAME argument is '?', the command presents a list of connection
      definitions from the connections file and prompts the user to
      select one, which is then displayed.

    * If the NAME argument is omitted, displays the current connection
      information if there is a current connection.

    Example showing a named connection definition:

    \b
      pywbemcli connection show svr1
        name: svr1
        ...

    Example for prompting for a connection definition:

    \b
      pywbemcli connection show ?
        0: svr1
        1: svr2
      Input integer between 0 and 2 or Ctrl-C to exit selection: : 0
        name: svr1
          ...
    """
    context.execute_cmd(lambda: cmd_connection_show(context, name, options))


@connection_group.command('delete', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@click.argument('name', type=str, metavar='NAME', required=False)
@add_options(help_option)
@click.pass_obj
def connection_delete(context, name):
    """
    Delete a WBEM connection definition.

    Delete a named connection definition from the connections file. If the NAME
    argument is omitted, prompt for selecting one of the connection definitions
    in the connections file.

    Example:

      pywbemcli connection delete blah
    """
    context.execute_cmd(lambda: cmd_connection_delete(context, name))


@connection_group.command('select', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@click.argument('name', type=str, metavar='NAME', required=False)
@click.option('-d', '--default', is_flag=True,
              default=False,
              help=u'If set, the connection is set to be the default '
                   u'connection in the connections file in addition to setting '
                   u'it as the current connection.')
@add_options(help_option)
@click.pass_obj
def connection_select(context, name, **options):
    """
    Select a WBEM connection definition as current or default.

    Select the connection definition named NAME from the connections file to
    be the current connection. The connection definition in the connections
    file must exist. If the NAME argument is omitted, a list of connection
    definitions from the connections file is presented with a prompt for the
    user to select a connection definition.

    If the --default option is set, the default connection is set to the
    selected connection definition, in addition.
    Once defined, the default connection will be used as a default in future
    executions of pywbemcli if none of the server-defining general options
    (i.e. --server, --mock-server, or --name) was used.

    The 'connection list' command marks the current connection with '*' and
    the default connection with '#'.

    Example of selecting a default connection in command mode:

    \b
      pywbemcli connection select myconn --default
      pywbemcli connection show
      name: myconn
        . . .

    Example of selecting just the current connection in interactive mode:

    \b
      pywbemcli
      pywbemcli> connection select myconn
      pywbemcli> connection show
      name: myconn
        . . .
    """

    context.execute_cmd(lambda: cmd_connection_select(context, name, options))


@connection_group.command('test', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@click.option('--test-pull', is_flag=True,
              default=False,
              help=u'If set, the connection is tested to determine if the'
                   u'DMTF defined pull operations (ex. OpenEnumerateInstances'
                   u'are implemented since these are optional.')
@add_options(help_option)
@click.pass_obj
def connection_test(context, **options):
    """
    Test the current connection with a predefined WBEM request.

    Execute the EnumerateClassNames operation on the default namespace against
    the current connection to confirm that the connection exists and is
    working.

    Examples:

      pywbemcli --name mysrv connection test
    """
    context.execute_cmd(lambda: cmd_connection_test(context, options))


@connection_group.command('save', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@click.argument('name', type=str, metavar='NAME', required=True)
@add_options(help_option)
@click.pass_obj
def connection_save(context, name):
    """
    Save the current connection to a new WBEM connection definition.

    Save the current connection to the connections file as a connection
    definition named NAME. The NAME argument is required.
    If a connection definition with that name already exists, it is
    overwritten without warning.

    In the interactive mode, general options that are connection related
    are applied to the current connection before it is saved.

    Examples:

      pywbemcli --server https://srv1 connection save mysrv
    """
    context.execute_cmd(lambda: cmd_connection_save(context, name))


@connection_group.command('list', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@click.option('-f', '--full', is_flag=True,
              default=False,
              help=u'If set, display the full table. Otherwise display '
                   u'a brief view(name, server, mock_server columns).')
@add_options(help_option)
@click.pass_obj
def connection_list(context, **options):
    """
    List the WBEM connection definitions.

    This command displays all entries in the connections file and the
    current connection if it exists and is not in the connections file as
    a table.

    \b
    '#' before the name indicates the default connection.
    '*' before the name indicates that it is the current connection.

    See also the 'connection select' command.
    """
    context.execute_cmd(lambda: cmd_connection_list(context, options))


################################################################
#
#   Common methods for The action functions for the connection click group
#
###############################################################


def export_statement(name, value):
    """ Export the name/value provided as:
        export name=value. We do not actually export but output the
        exprot definition.
    """
    if not isinstance(value, six.string_types):
        value = str(value)
    click.echo('export {}={}'.format(name, value))


def if_export_statement(name, value):
    """Export the statement if the value is not None"""
    if value is not None:
        export_statement(name, value)


def is_default_connection(context, connection):
    """
    Returns True if connection is the default connection. Where
    default connection is defined in repository.
    """

    default_connection = context.connections_repo.default_connection_name
    if default_connection and default_connection == connection.name:
        return True
    return False


def is_current_connection(context, connection):
    """Returns True if connection named name is the default connection"""

    if context.pywbem_server_exists():
        current_connection = context.pywbem_server
    else:
        current_connection = None
    if current_connection and current_connection.name == connection.name:
        return True
    return False


def show_connection_information(context, connection,
                                show_state=False, show_password=False):
    """
    Common function to display the connection information.  Note that this
    could also have been part of the context object since this is just the
    info from that object.
    """

    output_format = validate_output_format(context.output_format, 'TABLE')

    # Build state string, mock_server obj, password
    state_str = ''
    if show_state:
        state = []
        if is_current_connection(context, connection):
            state.append("current")
        if is_default_connection(context, connection):
            state.append("default")
        if state:
            state_str = ' ({})'.format(", ".join(state))

    if connection.password:
        disp_password = connection.password if show_password else '******'
    else:
        disp_password = connection.password

    headers = ["name", 'value  (state)']

    if connection.mock_server:
        mock_server = fold_strings(connection.mock_server, 20,
                                   fold_list_items=True,
                                   break_long_words=False,
                                   break_on_hyphens=False)
    else:
        mock_server = None

    if connection.ca_certs:
        ca_certs = fold_strings(connection.ca_certs, 20,
                                fold_list_items=True,
                                break_long_words=False)
    else:
        ca_certs = None

    rows = [['name', "{}{}".format(connection.name, state_str)],
            ['server', connection.server],
            ['default-namespace', connection.default_namespace],
            ['user', connection.user],
            ['password', disp_password],
            ['timeout', connection.timeout],
            ['use-pull', connection.use_pull],
            ['pull-max-cnt', connection.pull_max_cnt],
            ['verify', connection.verify],
            ['certfile', connection.certfile],
            ['keyfile', connection.keyfile],
            ['mock-server', mock_server],
            ['ca-certs', ca_certs]]

    context.spinner_stop()

    click.echo(format_table(rows, headers, title='Connection status:',
                            table_format=output_format))


def get_current_connection_name(context):
    """
    Return the name of the current connection or None if there is no
    current connection
    """
    return context.pywbem_server.name if context.pywbem_server_exists() \
        else None


def pick_connection(name, context, connections_repo):
    """
    If name is None use the interactive mode to select the connection from the
    list of connections in the connections file. If the name is provided, it is
    tested against the names in the connections file.  If it is not provided,

    Parameters:

      name (:term:`string):
        Name that will be validated against repository or Noe if the name
        is to be picked from a selection presented

      context (:class:`'~pywbemtools.pywbemcli._context_obj.ContextObj`):
        The current ContextObj instance.

      connections_repo (:class:`'~pywbemtools._context_obj.ContextObj`):

    Returns:
        name selected from connections  or None if no name selected.

    Raises

      click.ClickException: No connection file exists or if there is a name,
      parameter, that name does not exist in the connections file

    """
    context.spinner_stop()

    if not connections_repo.file_exists():
        raise click.ClickException(
            'Connections file "{0}" does not exist'.
            format(connections_repo.connections_file))

    if name:
        if name in connections_repo:
            return name
        raise click.ClickException(
            'Connection definition "{0}" not found in connections file "{1}"'.
            format(name, connections_repo.connections_file))

    conn_names = sorted(list(six.iterkeys(connections_repo)))
    return pick_one_from_list(context, conn_names,
                              "Select a connection or Ctrl-C to abort.")


def test_pull_operations(context, test_class):
    """
    Test if the pull operations are implemented by executing each of the
    commands and report results.  Uses test_class for the commands that
    require class input and an instance of that class for the commands that
    require an instance.

    Parameters:

      conn (:class:`~pywbem.WBEMConnection`):
        The connection object passed to the command that called this function.
        Used to execute commands on the WBEM server

      Returns:
        List tuples where each tuple consists of the name of the WBEM
        operation that was tested and the results of the test ('OK,
        'NOT SUPPORTED, or the exception if any exception other than
        CIMError(CIM_ERR_NOT_SUPPORTED)
    """

    def cimcall(request, *pargs, **kwargs):
        try:
            request(*pargs, **kwargs)
            return 'OK'

        except CIMError as er:
            if er.status_code == CIM_ERR_NOT_SUPPORTED:
                return 'NOT SUPPORTED'

            return er

    conn = context.pywbem_server.conn

    # Find a valid instance to use in tests.
    test_instances = conn.EnumerateInstanceNames(test_class)
    if not test_instances:
        raise click.ClickException("No instances found for test class {}".
                                   format(test_class))
    test_instance = test_instances[0]

    # execute each command and append the results to rows as a tuple
    rows = []
    result = cimcall(conn.OpenEnumerateInstances, test_class)
    rows.append(('OpenEnumerateInstances', result))

    result = cimcall(conn.OpenEnumerateInstancePaths, test_class)
    rows.append(('OpenEnumerateInstancePaths', result))

    result = cimcall(conn.OpenAssociatorInstances, test_instance)
    rows.append(('OpenAssociatorInstances', result))

    result = cimcall(conn.OpenAssociatorInstancePaths, test_instance)
    rows.append(('OpenAssociatorInstancePaths', result))

    result = cimcall(conn.OpenReferenceInstances, test_instance)
    rows.append(('OpenReferenceInstances', result))

    result = cimcall(conn.OpenReferenceInstancePaths, test_instance)
    rows.append(('OpenReferenceInstancePaths', result))

    result = cimcall(conn.OpenQueryInstances, "DMTF:CQL",
                     "SELECT * FROM CIM_ManagedElement")
    rows.append(('OpenQueryInstances', result))

    return rows


################################################################
#
#   Common methods for The action functions for the connection click group
#
###############################################################


def cmd_connection_export(context):
    """
        Save the current connection information to env variables using an
        export statement
    """
    context.spinner_stop()
    svr = context.pywbem_server
    if not svr:
        raise click.ClickException("No server currently defined as current")

    export_statement(PywbemServer.server_envvar, svr.server)

    if_export_statement(PywbemServer.defaultnamespace_envvar,
                        svr.default_namespace)
    if_export_statement(PywbemServer.user_envvar, svr.user)
    if_export_statement(PywbemServer.password_envvar, svr.password)
    if_export_statement(PywbemServer.verify_envvar, svr.verify)
    if_export_statement(PywbemServer.certfile_envvar, svr.certfile)
    if_export_statement(PywbemServer.keyfile_envvar, svr.keyfile)
    if_export_statement(PywbemServer.ca_certs_envvar, svr.ca_certs)
    if_export_statement(PywbemServer.timeout_envvar, svr.timeout)
    if_export_statement(PywbemServer.use_pull_envvar, svr.use_pull)
    # Additional exports available connections_file
    if_export_statement(PywbemServer.connections_file_envvar,
                        context.connections_repo.connections_file)


def cmd_connection_show(context, name, options):
    """
    Show the parameters that make up the current connection information if
    name is None. If Name exists, shows connections in connections file.
    If name is "?" present list of all connections for selection.
    """

    connections_repo = context.connections_repo

    current_name = get_current_connection_name(context)

    # If no name arg, use the current connection if it exists
    if not name:
        if not current_name:
            raise click.ClickException('No current connection exists.')
        name = current_name

    # ? means ask for connections file. However fallback to current connection
    # if there is no connections file and fail if no current connection.
    if name == '?':
        name = pick_connection(None, context, connections_repo)

    # Name is now defined
    if name == current_name:
        show_connection_information(context,
                                    context.pywbem_server,
                                    show_state=True,
                                    show_password=options['show_password'])
        return

    # Show a server definition from the connections file
    if connections_repo.file_exists():
        if name not in connections_repo:
            raise click.ClickException(
                'Connection definition "{0}" not found in connections file '
                '"{1}"'.format(name, connections_repo.connections_file))
        connection = connections_repo[name]
    else:   # no connections file
        raise click.ClickException(
            'Name: "{}" not current and no connections file {}'.format(
                name, connections_repo.connections_file))

    show_connection_information(context,
                                connection,
                                show_state=True,
                                show_password=options['show_password'])


def cmd_connection_test(context, options):
    """
    Test the current connection with a single command on the default_namespace.
    Uses enumerateClassNames against current workspace as most general
    possible operations that should work on all servers that support class
    operation. Even if class operations are not supported, a return from the
    server such as "unsupported" indicates the server exists.
    """
    conn = context.pywbem_server.conn
    output_format = validate_output_format(context.output_format, ['TABLE',
                                                                   'TEXT'])

    try:
        classnames = conn.EnumerateClassNames()
        context.spinner_stop()
    except CIMError as ce:
        # class request failed, try instance request
        try:
            test_classname = 'CIM_ManagedElement'
            if ce.status_code == 'CIM_ERR_NOT_SUPPORTED':
                conn.EnumerateInstances(test_classname)
        except Error as er:
            raise click.ClickException('Cannot interact with WBEM Server '
                                       '{}. EnumerateInstanceNames exception '
                                       '{} and EnumerateInstances {} exception '
                                       ' {} '.format(conn.url,
                                                     ce.status_code,
                                                     test_classname,
                                                     er))
    except Error as er:
        raise pywbem_error_exception(er)

    if classnames:
        if 'CIM_ManagedElement' in classnames:
            test_class = 'CIM_ManagedElement'
        else:
            test_class = classnames[0]
    else:
        raise click.ClickException('No classes found to test in namespace {}.'.
                                   format(conn.defaultnamespace))

    if options['test_pull']:
        result = test_pull_operations(context, test_class)

        if output_format_is_table(output_format):
            headers = ['Operation', 'Result']
            click.echo(format_table(
                result, headers,
                title='Pull Operation test results (Connection OK): {}'.
                format(conn.host),
                table_format=output_format))
        else:
            click.echo('Connection OK: {}'.format(conn.host))

            for row in result:
                click.echo("{}: {}".format(row[0], row[1]))

    else:
        click.echo('Connection OK: {}'.format(conn.host))


def cmd_connection_select(context, name, options):
    """
    Select an existing connection to use as the current WBEM server. This
    command accepts the click_context since it updates that context.

    If the --default flag is set, also set this connection as the persistent
    default connection.
    """
    connections_repo = context.connections_repo

    name = pick_connection(name, context, connections_repo)

    new_ctx = ContextObj(connections_repo[name],
                         context.output_format,
                         context.use_pull,
                         context.pull_max_cnt,
                         context.timestats,
                         context.log,
                         context.verbose,
                         context.pdb,
                         context.warn,
                         context.connections_repo,
                         context.interactive_mode,
                         False)

    # Update the root context making this context the basis for future
    # commands in the current interactive session
    ContextObj.update_root_click_context(new_ctx)

    # Close any existing connection.
    if context.is_connected():
        context.pywbem_server.disconnect()

    context.spinner_stop()
    if options['default']:
        connections_repo.default_connection_name = name
        click.echo('"{}" default and current'.format(name))
    else:
        click.echo('"{}" current'.format(name))


def cmd_connection_delete(context, name):
    """
    Delete a connection definition from the set of connections defined.

    This may be either defined by the optional name parameter or if there
    is no name provided, a select list will be presented for the user
    to select the connection to be deleted.
    """
    connections_repo = context.connections_repo

    # Select the connection with prompt if name is None.
    # This also stops the spinner
    name = pick_connection(name, context, connections_repo)

    current_name = get_current_connection_name(context)
    connections_repo.delete(name)

    default = 'default ' if current_name and current_name == name else ''
    click.echo('Deleted {} connection "{}".'.format(default, name))


def cmd_connection_save(context, name):
    """
    Saves the current connection definition with the name provided. A current
    connection must exist.
    The save function allows overwriting existing names
    """

    current_connection = context.pywbem_server
    if not current_connection:
        raise click.ClickException('No current connection connection. Use '
                                   '"connection list" to see more connection '
                                   'information')
    save_connection = deepcopy(current_connection)
    save_connection.name = name

    connections = context.connections_repo

    context.spinner_stop()
    try:
        connections.add(save_connection)
    except ConnectionsFileError as cfe:
        click.echo('Fatal error: {0}: {1}'.format(cfe.__class__.__name__, cfe),
                   err=True)
        raise click.Abort()


def cmd_connection_list(context, options):
    """
    List all of the current servers in the persistent repository line
    by line.  This method displays the information as a table independent
    of the value of the cmd line output_format general option.

    I includes both a full display that displays all variables and a brief
    display that attempts to keep the table in about 80 columns.
    """
    def build_row(options, name, svr):
        """
        Build a singe row of the table output and return it
        """
        if options['full']:
            return [name, svr.server, svr.default_namespace, svr.user,
                    svr.timeout, svr.use_pull, svr.pull_max_cnt, svr.verify,
                    svr.certfile, svr.keyfile, "\n".join(svr.mock_server)]
        return [name, svr.server, "\n".join(svr.mock_server)]

    connections_repo = context.connections_repo
    output_format = validate_output_format(context.output_format, 'TABLE')

    # build the table structure
    rows = []
    cur_sym = '*'  # single char representing current connection
    dflt_sym = '#'  # single char representing persisted default connection

    if connections_repo.file_exists():
        for name, svr in connections_repo.items():
            cc = cur_sym if is_current_connection(context, svr) else ''
            dc = dflt_sym if is_default_connection(context, svr) else ''
            name = '{}{}{}'.format(cc, dc, name)
            rows.append(build_row(options, name, svr))

    # add current connection if not in persistent connections
    if context.pywbem_server_exists():
        current_connection = context.pywbem_server
    else:
        current_connection = None

    if current_connection:
        cname = current_connection.name
        if connections_repo.file_exists():
            # Add connection that is current but not in connections repo
            if cname not in connections_repo:
                cname = '{}{}'.format('*', cname)
                rows.append(build_row(options, cname, current_connection))
        else:
            cname = '{}{}'.format('*', cname)
            rows.append(build_row(options, cname, current_connection))

    # NOTE: Does not show ca_certs because that creates a very big table
    # in particular if you use the default.
    if options['full']:
        headers = ['name', 'server', 'namespace', 'user',
                   'timeout', 'use-pull', 'pull-max-cnt', 'verify',
                   'certfile', 'keyfile', 'mock-server']
    else:
        headers = ['name', 'server', 'mock-server']

    context.spinner_stop()
    table_type = 'full' if options['full'] else 'brief'
    click.echo(format_table(
        sorted(rows),
        headers,
        title='WBEM server connections({0}): ({1}: default, {2}: '
              'current)\nfile: {3}'.format(
                  table_type, dflt_sym, cur_sym,
                  connections_repo.connections_file),
        table_format=output_format))



# (C) Copyright 2021 IBM Corp.
# (C) Copyright 2021 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click command definition for the statistic command group which includes
cmds for managing the gathering and display of statistics that pywbemcli
and the WBEM server can gather on the WBEM operations

NOTE: Commands are ordered in help display by their order in this file.
"""

from __future__ import absolute_import, print_function

import click
import six

from pywbem import Error, ValueMapping, CIMDateTime

from .pywbemcli import cli
from ._common import warning_msg, CMD_OPTS_TXT, GENERAL_OPTS_TXT, \
    validate_output_format, output_format_is_table, format_table, \
    SUBCMD_HELP_TXT
from ._common_options import add_options, help_option
from ._click_extensions import PywbemcliGroup, PywbemcliCommand

# NOTE: A number of the options use double-dash as the short form.  In those
# cases, a third definition of the options without the double-dash defines
# the corresponding option name, ex. 'include_qualifiers'. It should be
# defined with underscore and not dash

# Issue 224 - Exception in prompt-toolkit with python 2.7. Caused because
# with prompt-toolkit 2 + the completer requires unicode and click_repl not
# passing help as unicode in options as unicode
# NOTE: Insure that all option help attributes are unicode to get around this
#       issue


@cli.group('statistics', cls=PywbemcliGroup, options_metavar=GENERAL_OPTS_TXT,
           subcommand_metavar=SUBCMD_HELP_TXT)
@add_options(help_option)
def statistics_group():
    """
    Command group for WBEM operation statistics.

    This command group defines commands to control the gathering and display of
    statistical data about the WBEM operations issued by pywbemcli or processed
    by a WBEM server.

    Pywbemcli maintains statistics about the WBEM operations it has issued. This
    includes the Client Time measured on the client side and the Server Time
    measured by the WBEM server and reported in the 'WBEMServerResponseTime'
    header field in the CIM-XML response (if supported by the WBEM server).

    The WBEM server may also support maintaining statistics about the WBEM
    operations it has processed (possibly by multiple clients). Pywbemcli
    supports enabling or disabling the statistics gathering on the WBEM server
    via the 'GatherStatisticalData' property in the 'CIM_ObjectManager'
    instance for the WBEM server and supports retrieving and displaying the
    server maintained statistics.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'statistics' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@statistics_group.command('reset', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def statistics_reset(context):
    """
    Reset client maintained statistics.

    This command resets the counts in the statistics maintained by pywbemcli.
    This includes the server response times received from the WBEM server in
    the 'WBEMServerResponseTime' header field of the CIM-XML response,
    if supported and enabled.

    This command does not reset the server maintained statistics.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_statistics_reset(context))


@statistics_group.command('server-on', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def statistics_server_on(context):
    """
    Enable server maintained statistics.

    This command activates the gathering of statistics in the current
    WBEM server and also activates the returning of server response times in
    the CIM-XML response for inclusion in the client maintained statistics,
    by setting the 'GatherStatisticalData' property to True in the
    'CIM_ObjectManager' instance for the WBEM server.

    This command fails if the server does not support gathering statistics
    or does not allow a client to modify the state of statistics gathering.

    This command does not affect the state of the client maintained statistics
    other than whether the server response times are included. See the
    '--timestats' general option for controlling the displaying of client
    maintained statistics.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_statistics_server_on(context))


@statistics_group.command('server-off', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def statistics_server_off(context):
    """
    Disable server maintained statistics.

    This command deactivates the gathering of statistics in the current
    WBEM server and also deactivates the returning of server response times in
    the CIM-XML response for inclusion in the client maintained statistics,
    by setting the 'GatherStatisticalData' property to False in the
    'CIM_ObjectManager' instance for the WBEM server.

    This command fails if the server does not support gathering statistics
    or does not allow a client to modify the state of statistics gathering.

    This command does not affect the state of the client maintained statistics
    other than whether the server response times are included. See the
    '--timestats' general option for controlling the displaying of client
    maintained statistics.
    """
    # pylint: disable=too-many-function-args

    context.execute_cmd(lambda: cmd_statistics_server_off(context))


@statistics_group.command('server-show', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def statistics_server_show(context):
    """
    Display server maintained statistics.

    Retrieve and display the statistics gathered by the current WBEM server.
    This requires statistics gathering to be enabled on the server (see
    'statistics server-on' command).

    This command fails if the server does not support gathering statistics.

    These statistics are independent of the client maintained statistics which
    can be displayed with the command 'statistics show'.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_statistics_server_show(context))


@statistics_group.command('show', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def statistics_show(context):
    """
    Display client maintained statistics.

    Display the statistics gathered by pywbemcli.
    This includes the server response times received from the WBEM server in
    the 'WBEMServerResponseTime' header field of the CIM-XML response,
    if supported and enabled.

    These statistics are independent of the server maintained statistics which
    can be displayed with the command 'statistics server-show'.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_statistics_show(context))


@statistics_group.command('status', cls=PywbemcliCommand,
                          options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def statistics_status(context):
    """
    Show enabled status of client and server maintained statistics.

    Show the enabled status for displaying the statistics gathered by
    pywbemcli, and for gathering the statistics on the current WBEM server.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_statistics_status(context))


###############################################################
#         Server cmds  Common components
###############################################################

OBJMGR_CLN = "CIM_ObjectManager"
OBJMGR_STAT_PROPERTY_NAME = "GatherStatisticalData"
CIMOM_STATISTICAL_DATA_CLASS = "CIM_CIMOMStatisticalData"


def get_objmgr_inst(context):
    """
    Return the state of statistics gathering on the server.
    """
    conn = context.pywbem_server.conn
    wbem_server = context.pywbem_server.wbem_server
    try:
        interop_ns = wbem_server.interop_ns
    except Error as er:
        raise click.ClickException(
            'Cannot access interop namespace. Exception: {}'.format(er))

    try:
        objmanager_insts = conn.EnumerateInstances(OBJMGR_CLN,
                                                   namespace=interop_ns)

        # There must be only one CIM_ObjectManager instance. None causes
        # failure and more than 1 causes warning
        if not objmanager_insts:
            raise click.ClickException(
                'No instances of class {} found on server'.format(OBJMGR_CLN))

        if len(objmanager_insts) > 1:
            warning_msg(
                "Server returned multiple {} instances. Using first.".
                format(objmanager_insts[0].classname))

    except Error as er:
        raise click.ClickException(
            'Failure trying to access object manager instances of class {} '
            'namespace {}. Error: {}'.
            format(OBJMGR_CLN, interop_ns, er))
    objmgr_inst = objmanager_insts[0]

    return objmgr_inst


def to_on_off(state):
    """
    Convert boolean to "on" or "off.

    Return string
        String containing "on" or "off"
    """
    return "on" if state else "off"


def set_server_statistics(context, desired_state):
    """
    Set the server statistics state to on or off, enabling or disabling
    the statistics.

    Parameters:
      desired_state (:class": `boolean)
        Value desired for the GatherStatisticalData property
    """
    conn = context.pywbem_server.conn
    wbem_server = context.pywbem_server.wbem_server

    interop_ns = wbem_server.interop_ns

    objmgr = get_objmgr_inst(context)

    context.spinner_stop()

    try:
        cur_state = objmgr[OBJMGR_STAT_PROPERTY_NAME]
    except KeyError as ke:
        raise click.ClickException(
            'Failure accessing class {} property {}. namespace {}:Error: {}'.
            format(OBJMGR_CLN, OBJMGR_STAT_PROPERTY_NAME, interop_ns, ke))

    if cur_state == desired_state:
        click.echo("Server statistics gathering was already {}".
                   format(to_on_off(desired_state)))
        return

    objmgr[OBJMGR_STAT_PROPERTY_NAME] = desired_state

    # Otherwise modify the property in the server
    try:
        conn.ModifyInstance(objmgr, PropertyList=[OBJMGR_STAT_PROPERTY_NAME],
                            IncludeQualifiers=False)
    except Error as ce:
        raise click.ClickException('Modification of GatherStatisticalData '
                                   'property in instance {} on server failed: '
                                   'Exception {}'.format(objmgr.path, ce))

    modified_objmgr = conn.GetInstance(objmgr.path)

    if modified_objmgr[OBJMGR_STAT_PROPERTY_NAME] != desired_state:
        raise click.ClickException(
            'The WBEM server reported success when changing the {} property '
            'in its {} instance to {} but did not actually change it.'.
            format(OBJMGR_STAT_PROPERTY_NAME, OBJMGR_CLN, interop_ns))

    click.echo("Server statistics gathering set to {}".
               format(to_on_off(desired_state)))


###############################################################
#         Statistics cmds action functions
###############################################################


def cmd_statistics_server_on(context):
    """
    Attempt to enable the statistics gathering on the current server if there
    is a current server defined.
    """
    set_server_statistics(context, True)


def cmd_statistics_server_off(context):
    """
    Attempt to disable the statistics gathering on the current server if there
    is a current server defined.
    """
    set_server_statistics(context, False)


def cmd_statistics_status(context):
    """
    Show current status of both client and server statistics if there
    is a connection.

    It gets the status of the server statistics from the WBEM server
    and the client statistics from the context and displays either a
    table or text representation of the status.

    If there is an issue with getting statistics information from the server,
    it displays detailed information about the server response.

    """

    # validate output format with default of table
    output_format = validate_output_format(context.output_format,
                                           ['TABLE', 'TEXT'])

    # Display server status as on/off
    try:
        objmgr = get_objmgr_inst(context)
        try:
            svr_status = objmgr[OBJMGR_STAT_PROPERTY_NAME]
            svr_status = to_on_off(svr_status)
        except KeyError:
            svr_status = "No property {}".format(OBJMGR_STAT_PROPERTY_NAME)
    except Error as er:
        svr_status = "Not settable {}".format(er)

    context.spinner_stop()

    pywbemcli_stats_status = to_on_off(context.timestats)
    if output_format_is_table(output_format):
        headers = ['Item', 'Status']
        rows = [['client statistics display', pywbemcli_stats_status],
                ['server statistics gathering', svr_status]]
        click.echo(format_table(rows, headers, "Statistics status"))
    else:
        click.echo("Statistics status: client={}; server={}".
                   format(pywbemcli_stats_status, svr_status))


def cmd_statistics_reset(context):
    """
    Reset the client statistics. This command does not impact the
    statistics retrieved from the server with ``statistics server-show``.
    It does reset the statistics kept in the pywbem client for server
    response time.
    """

    conn = context.pywbem_server.conn
    conn.statistics.reset()

    context.spinner_stop()
    click.echo("Pywbemcli statistics reset")


def cmd_statistics_show(context):
    """
    If the statistics are enabled for the client and the connection exits,
    display the statistics
    """

    conn = context.pywbem_server.conn

    context.spinner_stop()
    click.echo(context.format_statistics(conn.statistics, context))


def cmd_statistics_server_show(context):
    """
    If server statistics are enabled get them from the server and display.

    This command may fail if the WBEM server has not implemented the
    management of detailed statistics.
    """
    conn = context.pywbem_server.conn

    def avg_value(value_, num_ops):
        """
        Compute the average value from the value_ input and the number of
        operations.
        If the number of operations is 0, return 0

        Parameters:
          value_ (:class:`int` or :class:`py:float`):

          num_ops (:class:`int`):
            The number of operations executed

        Returns:
          Numeric containing the value / num_ops or
          0 if num_ops is 0.
        """
        if num_ops:
            return value_ / num_ops

        assert value_ == 0
        return 0

    def avg_time_ms(total_time, num_ops):
        """
        Return the average elapsed time of the operations in milliseconds.

        Parameters:
          total_time (:class:`CIMDateTime`):
            The total elapsed time of the operations, as an interval.

          num_ops (:class:`int`):
            The number of operations executed.

          Returns:
            :class:`int`: The average elapsed time of the operations in
            milliseconds.
        """
        assert isinstance(total_time, CIMDateTime)
        assert total_time.is_interval
        milliseconds = total_time.timedelta.total_seconds() * 1000
        avg = avg_value(milliseconds, num_ops)
        return avg

    conn = context.pywbem_server.conn
    wbem_server = context.pywbem_server.wbem_server
    output_fmt = validate_output_format(context.output_format, 'TABLE')

    objmgr = get_objmgr_inst(context)
    try:
        svr_status = objmgr[OBJMGR_STAT_PROPERTY_NAME]
    except KeyError:
        svr_status = "No property {}".format(OBJMGR_STAT_PROPERTY_NAME)
    except Error as er:
        svr_status = "Not settable {}".format(er)

    if isinstance(svr_status, six.string_types):
        raise click.ClickException("ObjectManager config setting not found {}".
                                   format(svr_status))

    if not svr_status:
        raise click.ClickException("WBEM server CIM_ObjectManager "
                                   "statistics off")

    # We must find the
    # namespace containing statistical data since apparently not clearly
    # documented in specs. Ex. Pegasus has in in root/cimv2 in test build
    namespaces = wbem_server.namespaces
    results = None
    stats_namespace = None
    for ns in namespaces:
        try:
            _ = conn.GetClass(CIMOM_STATISTICAL_DATA_CLASS, namespace=ns)
        except Error:
            continue
        try:
            results = conn.PyWbemcliEnumerateInstances(
                ClassName=CIMOM_STATISTICAL_DATA_CLASS,
                namespace=ns,
                LocalOnly=False,
                IncludeQualifiers=False,
                DeepInheritance=True,
                IncludeClassOrigin=False)
            stats_namespace = ns
            if results:
                break
        except Error:
            continue

    if not results:
        raise click.ClickException(
            "No WBEM server class {} or class returns no data on "
            "WBEM Server".
            format(CIMOM_STATISTICAL_DATA_CLASS))

    if context.verbose:
        click.echo("{} returns instances in namespace {}".
                   format(CIMOM_STATISTICAL_DATA_CLASS, stats_namespace))

    try:
        op_type_vm = ValueMapping.for_property(conn, stats_namespace,
                                               CIMOM_STATISTICAL_DATA_CLASS,
                                               'OperationType')

        results = conn.PyWbemcliEnumerateInstances(
            ClassName=CIMOM_STATISTICAL_DATA_CLASS,
            namespace=stats_namespace,
            LocalOnly=False,
            IncludeQualifiers=False,
            DeepInheritance=True,
            IncludeClassOrigin=False)
        if not results:
            raise click.ClickException(
                "WBEM server Class {} returns no data".
                format(CIMOM_STATISTICAL_DATA_CLASS))

        row_tuples = []
        for inst in results:
            # Only show lines that have a non-zero number of operations
            if inst['NumberOfOperations'] == 0:
                continue
            op_type_name = op_type_vm.tovalues(inst['OperationType'])
            # If op_type_name is other, get op_type_name from other property
            if op_type_name.lower() == 'other':
                op_type_name = inst['OtherOperationType']
            number_of_operations = inst['NumberOfOperations']

            avg_provider_time_ms = avg_time_ms(
                inst['ProviderElapsedTime'], number_of_operations)
            avg_cimom_time_ms = avg_time_ms(
                inst['CimomElapsedTime'], number_of_operations)
            # Note: The CimomElapsedTime property is explicitly described
            #       to be the time just in the CIMOM without provider time.
            #       Since we want to be able to relate that to the server time
            #       in the client statistics, we calculate the server time
            #       as CIMOM time plus provider time.
            avg_server_time_ms = avg_cimom_time_ms + avg_provider_time_ms

            avg_request_size = avg_value(inst['RequestSize'],
                                         number_of_operations)
            avg_response_size = avg_value(inst['ResponseSize'],
                                          number_of_operations)

            row_tuples.append(
                (op_type_name, number_of_operations, avg_server_time_ms,
                 avg_provider_time_ms, avg_request_size, avg_response_size))

        # Index of the items in the row tuples
        operation_idx = 0
        count_idx = 1
        server_time_idx = 2
        provider_time_idx = 3
        request_size_idx = 4
        response_size_idx = 5

        # Sort the rows by descending server time.
        # Note: Because the server time first needs to be calculated which may
        #       be WBEM server dependent, the sorting is performed as a
        #       separate step.
        rows = []
        for row_tuple in sorted(row_tuples,
                                key=lambda tup: tup[server_time_idx],
                                reverse=True):
            rows.append(
                (row_tuple[operation_idx], row_tuple[count_idx],
                 "{:.3f}".format(row_tuple[server_time_idx]),
                 "{:.3f}".format(row_tuple[provider_time_idx]),
                 "{:.0f}".format(row_tuple[request_size_idx]),
                 "{:.0f}".format(row_tuple[response_size_idx])))

        headers = ['Operation', 'Count', 'Server Time\n[ms]',
                   'Provider Time\n[ms]', 'Request Size\n[B]',
                   'Response Size\n[B]']

        context.spinner_stop()
        click.echo(format_table(
            rows, headers,
            title='Server statistics',
            table_format=output_fmt))
    except Error as er:
        raise click.ClickException("Statistics retrieval from WBEM server "
                                   "failed. Exception {}".
                                   format(er))



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility Functions applicable across multiple components of pywbemcli.
"""

from __future__ import print_function, absolute_import

import os
import warnings
import mock
import six
import click

from . import config

__all__ = []

# Env var for overriding the terminal width
TERMWIDTH_ENVVAR = 'PYWBEMCLI_TERMWIDTH'

# Keep the following connections file definitions in sync with help text of
# the "--connections-file" option in pywbemcli.py and generaloptions.rst, and
# with the use of the base file name in several other .rst and .py files.

# Base file name of the connections file
# The B08_* file name was used before pywbemcli 0.8.
CONNECTIONS_FILENAME = '.pywbemcli_connections.yaml'
B08_CONNECTIONS_FILENAME = 'pywbemcli_connection_definitions.yaml'

# Path name of default connections file directory.
DEFAULT_CONNECTIONS_DIR = os.path.expanduser("~")

# Path name of default connections file
# The B08_* path name was used before pywbemcli 0.8.
DEFAULT_CONNECTIONS_FILE = os.path.join(DEFAULT_CONNECTIONS_DIR,
                                        CONNECTIONS_FILENAME)
B08_DEFAULT_CONNECTIONS_FILE = os.path.join(DEFAULT_CONNECTIONS_DIR,
                                            B08_CONNECTIONS_FILENAME)


def ensure_bytes(obj):
    """
    If the input object is a string, make sure it is returned as a Byte string,
    as follows:

    * If the input object already is a Byte string, it is returned unchanged.
    * If the input object is a Unicode string, it is converted to a Byte string
      using the UTF-8 encoding.
    * Otherwise, the input object was not a string and is returned unchanged.
    """
    if isinstance(obj, six.text_type):
        return obj.encode("utf-8")
    return obj


def ensure_unicode(obj):
    """
    If the input object is a string, make sure it is returned as a Unicode
    string, as follows:

    * If the input object already is a Unicode string, it is returned unchanged.
    * If the input object is a Byte string, it is converted to a Unicode string
      using the UTF-8 encoding.
    * Otherwise, the input object was not a string and is returned unchanged.
    """
    if isinstance(obj, six.binary_type):
        return obj.decode("utf-8")
    return obj


def to_unicode(obj):
    """
    Convert the input Byte string to a Unicode string.

    The input object must be a Byte string.
    """
    return obj.decode("utf-8")


def _formatwarning(message, category, filename, lineno, line=None):
    # pylint: disable=unused-argument
    """
    Replacement for warnings.formatwarning() that is monkey patched in.
    """
    return "{}: {}\n".format(category.__name__, message)


def pywbemcliwarn(*args, **kwargs):
    """
    Pywbemcli monkey patch for the warnings.warn function; substitutes our
    own formatter.
    """
    with mock.patch.object(warnings, 'formatwarning', _formatwarning):
        warnings.warn(*args, **kwargs)


def pywbemcliwarn_explicit(*args, **kwargs):
    """
    Pywbemcli monkey patch for the warnings.warn_explicit function;
    substitutes our own formatter
    """
    with mock.patch.object(warnings, 'formatwarning', _formatwarning):
        warnings.warn_explicit(*args, **kwargs)


def get_terminal_width():
    """
    Return the terminal width to use.

    Note: On Windows, click.get_terminal_size() results in terminal sizes of 0
    in some cases.
    """

    terminal_width = os.getenv(TERMWIDTH_ENVVAR, None)
    if terminal_width:
        try:
            terminal_width = int(terminal_width)
            return terminal_width
        except ValueError:
            pass

    if config.USE_TERMINAL_WIDTH:
        return click.get_terminal_size()[0]

    return config.DEFAULT_TABLE_WIDTH



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions to persist and restore the connections table.  This works from
a file that maintains defined connections. If any exist they are
loaded at startup and available through an interactive session.  Methods
are provided to create, delete, and view existing connections. A set method
allows setting the current active connection into the repository.
"""

from __future__ import absolute_import, print_function

import os
from contextlib import contextmanager
import yaml
import six
import yamlloader
import click

from ._pywbem_server import PywbemServer
from ._pywbemcli_operations import delete_mock_cache
from ._utils import DEFAULT_CONNECTIONS_FILE, B08_DEFAULT_CONNECTIONS_FILE

if six.PY2:
    import codecs  # pylint: disable=wrong-import-order

BAK_FILE_SUFFIX = 'bak'


class ConnectionsFileError(Exception):
    """
    Base class for connections file related exceptions. Exceptions are not
    raised using this class, but this class can be used to catch all of its
    derived exceptions.
    """
    def __init__(self, message):
        """
        Parameters:

          message (:term:`string`): Message text for the exception.
        """
        # pylint: disable=useless-super-delegation
        super(ConnectionsFileError, self).__init__(message)


class ConnectionsFileWriteError(ConnectionsFileError):
    """
    Exception indicating an error when writing the connections file.

    This includes I/O and OS errors during writing of the file or during
    renaming of temporary or backup files.
    """
    def __init__(self, connections_file, message):
        """
        Parameters:

          connections_file (:term:`string`): Path name of connections file.

          message (:term:`string`): Text describing just the error
            (without text like 'Cannot write connection file {file}').
        """
        msg = 'Cannot write connections file "{0}": {1}'. \
            format(connections_file, message)
        super(ConnectionsFileWriteError, self).__init__(msg)


class ConnectionsFileLoadError(ConnectionsFileError):
    """
    Exception indicating an error when loading the connections file.

    This includes errors due to invalid YAML format at the syntax or logical
    level, missing required properties, I/O and OS errors during loading of the
    file or during renaming of the file when migrating an old file.
    """
    def __init__(self, connections_file, message):
        """
        Parameters:

          connections_file (:term:`string`): Path name of connections file.

          message (:term:`string`): Text describing just the error
            (without text like 'Cannot load connection file {file}').
        """
        msg = 'Cannot load connections file "{0}": {1}'.format(connections_file,
                                                               message)
        super(ConnectionsFileLoadError, self).__init__(msg)


@contextmanager
def open_text_file(filename, file_mode):
    """
    Context manager that opens the specified file in text mode with UTF-8
    encoding and closes it upon exit.

    This method encapsulates a difference between Python 2 and Python 3
    when opening files in text mode with UTF-8 encoding.

    Parameters:

      filename (:term:`string`):
        Path name of the file to be opened.

      file_mode (:term:`string`):
        File mode, for example:
        - 'r' opens the file for reading.
        - 'w' opens the file for writing and truncates an existing file.
        Must be a text mode, i.e. 'b' is not permitted in the file mode.

    Returns:

      File-like object.

    Raises:
      OSError
      IOError (on Python 2, and possibly for lower level errors)
    """
    assert 'b' not in file_mode
    if six.PY2:
        # pylint: disable=consider-using-with
        fp = codecs.open(filename, mode=file_mode, encoding='utf-8')
    else:
        # pylint: disable=consider-using-with
        fp = open(filename, file_mode, encoding='utf8')
    yield fp
    fp.close()


class ConnectionRepository(object):
    # pylint: disable=useless-object-inheritance
    """
    A connection repository that contains the named connection definitions
    and provides methods to manage these connection definitions.

    The connection definitions are automatically persisted in a
    :term:`connections file`.

    This class loads the connections file in a lazy manner upon access by one
    of the methods that need it to be loaded. When modifications are made,
    the connections file is immediately written back, creating a backup file
    of the current connections file.

    Each connection definition is represented as a PywbemServer object.

    This class supports the dictionary interface for accessing the connection
    definitions, using the connection name as a key, and the PywbemServer object
    as a value. The dictionary interface supports a subset of read-only
    operations.

    The connections file is a YAML file, with a format as shown in the following
    example that defines a mock connection named 'mock1', a server connection
    named 'server1', and defines 'mock1' as the default connection. The
    properties of each connection definition are the attributes of PywbemServer
    objects:

        connection_definitions:
            mock1:
                name: mock1
                server: null
                user: null
                password: null
                default-namespace: root/cimv2
                timeout: 30
                use_pull: null
                pull_max_cnt: null
                verify: true
                certfile: null
                keyfile: null
                ca-certs: null
                mock-server:
                - tests/unit/simple_mock_model.mof
                - tests/unit/simple_mock_invokemethod_v1old.py
            server1:
                name: server1
                server: https://woot.com
                user: foo
                password: pass
                default-namespace: root/cimv2
                timeout: 30
                use_pull: null
                pull_max_cnt: null
                verify: false
                certfile: null
                keyfile: null
                ca-certs: null
                mock-server: []
        default_connection_name: mock1
    """

    # Name of YAML property in the connections file that contains the
    # named connection definitions
    connections_group_name = 'connection_definitions'

    # Name of YAML property in the connections file that contains the
    # default connection name
    default_connection_grp_name = 'default_connection_name'

    def __init__(self, connections_file, verbose=None):
        """
        Initialize the object. The connections file is not yet loaded.

        Parameters:

          connections_file (:term:`string`):
            Path name of the connections file. Must not be `None`.

          verbose (:class:`py:bool`):
            If `True` enables progress console displays.
        """

        # Dictionary of connection definitions.
        # Key: Name of the connection definition.
        # Value: PywbemServer object representing the connection definition.
        self._pywbemcli_servers = {}

        # Flag indicating whether the connections file has been loaded.
        self._loaded = False

        # Path name of the connections file.
        self._connections_file = connections_file

        # Name of the default connection definition.
        # `None`, if there is no default connection definition.
        self._default_connection_name = None
        self._verbose = verbose

    @property
    def connections_file(self):
        """
        :term:`string`: Path name of the connections file. Will not be `None`.
        """
        return self._connections_file

    @property
    def default_connection_name(self):
        """
        :term:`string`: Name of the default connection definition. If `None`,
        there is no default connection definition.

        Loads the connections file, if not yet loaded.

        The default connection is selected on pywbemcli startup if no
        connection is specified.

        Raises:
          ConnectionsFileLoadError
        """
        self._load_connections_file()
        return self._default_connection_name

    @default_connection_name.setter
    def default_connection_name(self, name):
        """
        Set the default connection definition.

        Loads the connections file, if not yet loaded, and updates it.

        Parameters:

          name (:term:`string` or `None`):
            The name of an existing connection definition to be set as the
            default connection definition.
            `None` will unset the default connection definition.

        Raises:
          ConnectionsFileLoadError
          ConnectionsFileWriteError
          KeyError: Connection definition not found
        """
        self._load_connections_file()
        if name is None or name in self._pywbemcli_servers:
            self._default_connection_name = name
            self._write_connections_file()
        else:
            raise KeyError('Connection definition {!r} not found'.
                           format(name))

    def file_exists(self):
        """
        Test if the connections file exists.

        Returns:
            :class:`py:bool`: `True` if the connections file exists and `False`
            if it does not exist.
        """
        return os.path.isfile(self._connections_file)

    def __str__(self):
        """
        Return a one-line human readable string that contains path name and
        existence status of the connections file, and count of connection
        definitions in this object.

        Loads the connections file, if not yet loaded.

        Raises:
          ConnectionsFileLoadError
        """
        exists = 'exists' if self.file_exists() else "does not exist"
        self._load_connections_file()
        count = len(self)
        return ('Connection repository with {0} connection definitions and '
                'connections file "{1}" ({2})'.
                format(count, self._connections_file, exists))

    def __repr__(self):
        """
        Return a string representation of the connections dictionary in this
        object that is suitable for debugging and other key parameters.

        Loads the connections file, if not yet loaded.

        Raises:
          ConnectionsFileLoadError
        """
        exists = 'exists' if self.file_exists() else "does not exist"
        self._load_connections_file()
        items = ["{0!r}: {1}".format(k, v)
                 for k, v in six.iteritems(self._pywbemcli_servers)]
        items_str = ', '.join(items)
        return ("{0}({{{1}}}, connections_file={2!r} ({3}), loaded={4!r})".
                format(self.__class__.__name__, items_str,
                       self._connections_file, exists, self._loaded))

    def __contains__(self, name):
        """
        Return a boolean indicating whether the specified connection definition
        exists.

        Loads the connections file, if not yet loaded.

        Parameters:

          name (:term:`string`): Name of the connection definition.

        Raises:
          ConnectionsFileLoadError
        """
        self._load_connections_file()
        return name in self._pywbemcli_servers

    def __getitem__(self, name):
        """
        Return the PywbemServer object for the specified connection definition.

        Loads the connections file, if not yet loaded.

        Parameters:

          name (:term:`string`): Name of the connection definition.

        Raises:
          ConnectionsFileLoadError
          KeyError: Connection definition not found
        """
        self._load_connections_file()
        return self._pywbemcli_servers[name]

    def __len__(self):
        """
        Return the count of connection definitions.

        Loads the connections file, if not yet loaded.

        Raises:
          ConnectionsFileLoadError
        """
        self._load_connections_file()
        return len(self._pywbemcli_servers)

    def __iter__(self):
        """
        Return an iterator through the names of the connection definitions.

        Loads the connections file, if not yet loaded.

        Raises:
          ConnectionsFileLoadError
        """
        self._load_connections_file()
        return six.iterkeys(self._pywbemcli_servers)

    def items(self):
        """
        Return a list of the connection definitions, as tuple(name, conn)
        where conn is a PywbemServer object.

        Loads the connections file, if not yet loaded.

        Raises:
          ConnectionsFileLoadError
        """
        self._load_connections_file()
        return list(self.iteritems())

    def keys(self):
        """
        Return a list of the connection names.

        Loads the connections file, if not yet loaded.

        Raises:
          ConnectionsFileLoadError
        """
        self._load_connections_file()
        return list(self.iterkeys())

    def iteritems(self):
        """
        Return an iterator through the connection definitions, as
        tuple(name, conn) where conn is a PywbemServer object.

        Loads the connections file, if not yet loaded.

        Raises:
          ConnectionsFileLoadError
        """
        self._load_connections_file()
        return six.iteritems(self._pywbemcli_servers)

    def iterkeys(self):
        """
        Return an iterator through the connection names.

        Loads the connections file, if not yet loaded.

        Raises:
          ConnectionsFileLoadError
        """
        self._load_connections_file()
        return six.iterkeys(self._pywbemcli_servers)

    def _load_connections_file(self):
        """
        Load the connections file into the connection repository, if not yet
        loaded.

        Repeated calls to this method will load the file only once.

        Old connections files are migrated if needed, and a message about that
        is issued to stdout.

        Raises:
          ConnectionsFileLoadError
        """
        if self._loaded:
            return

        # Migrate an old connections file
        if self._connections_file == DEFAULT_CONNECTIONS_FILE and \
                not os.path.isfile(self._connections_file) and \
                os.path.isfile(B08_DEFAULT_CONNECTIONS_FILE):

            try:
                os.rename(B08_DEFAULT_CONNECTIONS_FILE,
                          DEFAULT_CONNECTIONS_FILE)
            except (OSError, IOError) as exc:
                raise ConnectionsFileLoadError(
                    self._connections_file,
                    'Error migrating old connections file "{0}": {1}'.
                    format(B08_DEFAULT_CONNECTIONS_FILE, exc))

            click.echo("Migrated old connections file {0!r} to {1!r}".
                       format(B08_DEFAULT_CONNECTIONS_FILE,
                              DEFAULT_CONNECTIONS_FILE))

        # If the file does not exist, the connection repo still has the
        # initial state at this point, and it remains empty.
        if not os.path.isfile(self._connections_file):
            return

        # Load the existing file.
        try:
            with open_text_file(self._connections_file, 'r') as _fp:
                try:
                    dict_ = yaml.safe_load(_fp)
                except (TypeError, yaml.YAMLError) as exc:
                    raise ConnectionsFileLoadError(
                        self._connections_file,
                        'Invalid YAML syntax: {0}'.format(exc))

                # Try building dictionary of server definitions
                try:
                    connections_dict = \
                        dict_[ConnectionRepository.connections_group_name]
                except TypeError as te:
                    raise ConnectionsFileLoadError(
                        self._connections_file,
                        'Invalid type of YAML property {0}: {1}'.
                        format(ConnectionRepository.connections_group_name, te))
                except KeyError:
                    raise ConnectionsFileLoadError(
                        self._connections_file,
                        'Missing YAML property {0}'.
                        format(ConnectionRepository.connections_group_name))

                # Try getting the default connection name
                try:
                    self._default_connection_name = dict_[
                        ConnectionRepository.default_connection_grp_name]
                except KeyError:
                    raise ConnectionsFileLoadError(
                        self._connections_file,
                        'Missing YAML property {0}'.
                        format(
                            ConnectionRepository.default_connection_grp_name))

                # Build the PywbemServer object for each connection definition
                for name, svr in six.iteritems(connections_dict):
                    try:
                        server = PywbemServer.create(
                            replace_underscores=True,
                            connections_file=self._connections_file,
                            **svr)
                    except TypeError as te:
                        raise ConnectionsFileLoadError(
                            self._connections_file,
                            'Invalid attribute type in connection '
                            'definition "{0}": {1}'.format(name, te))
                    except ValueError as ve:
                        raise ConnectionsFileLoadError(
                            self._connections_file,
                            'Invalid attribute value in connection '
                            'definition "{0}": {1}'.format(name, ve))
                    self._pywbemcli_servers[name] = server
                self._loaded = True
                if self._verbose:
                    click.echo("Connections file loaded: {}".
                               format(self._connections_file))

        except (OSError, IOError) as exc:
            raise ConnectionsFileLoadError(
                self._connections_file,
                'Error opening the file: {0}'.format(exc))

    def add(self, svr_definition):
        """
        Add a connection definition to the connections repository.

        If the connection definition already exists, it is replaced. If that is
        undesired, the existence of the connection definition can be checked
        before calling this method.

        Loads the connections file, if not yet loaded, and updates it.

        Parameters:

          svr_definition (:class:`PywbemServer`):
            An instance of PywbemServer representing the connection definition.
            The PywbemServer object must have set its `name` attribute.

        Raises:
          ConnectionsFileLoadError
          ConnectionsFileWriteError
        """
        assert isinstance(svr_definition.name, six.string_types)

        # If the file exists load it. Otherwise the write will create it
        if os.path.isfile(self._connections_file):
            self._load_connections_file()

        self._pywbemcli_servers[svr_definition.name] = svr_definition
        self._write_connections_file()

    def delete(self, name):
        """
        Delete a connection definition from the connections repository.

        Loads the connections file, if not yet loaded, and updates it.

        Parameters:

          name (:term:`string`):
            Name of the connection definition to delete.

        Raises:
          ConnectionsFileLoadError
          ConnectionsFileWriteError
          KeyError: Connection definition not found
        """
        self._load_connections_file()

        # Delete mock cache if it exists
        delete_mock_cache(self._connections_file, name)

        # Delete connection definition
        del self._pywbemcli_servers[name]

        # Unset default connection if it is the one being deleted
        if name == self._default_connection_name:
            self._default_connection_name = None

        self._write_connections_file()

    def _write_connections_file(self):
        """
        Write the connection repository state to the connections file.

        If there is an existing connections file, it is backed up with a suffix
        of '.bak' before writing the connections file.

        Raises:
          ConnectionsFileWriteError
        """
        conn_dict = {}
        if self._pywbemcli_servers:
            conn_dict = {name: value.to_dict() for name, value in
                         self._pywbemcli_servers.items()}

            # build dictionary for YAML output
            yaml_dict = {
                ConnectionRepository.connections_group_name: conn_dict,
                ConnectionRepository.default_connection_grp_name:
                self._default_connection_name
            }

            # Write to tmpfile and if successful create backup file and
            # move the tmpfile to be the new connections file contents.
            tmpfile = '{}.tmp'.format(self._connections_file)

            try:
                with open_text_file(tmpfile, 'w') as _fp:
                    data = yaml.dump(yaml_dict,
                                     encoding=None,
                                     allow_unicode=True,
                                     default_flow_style=False,
                                     indent=4,
                                     Dumper=yamlloader.ordereddict.CSafeDumper)
                    data = data.replace('\n\n', '\n')  # YAML dump dups newlines
                    _fp.write(data)
                    # Data gets flushed to disk when closing the file upon exit
                    # from the 'with' clause.
            except (OSError, IOError) as exc:
                raise ConnectionsFileWriteError(
                    self._connections_file,
                    'Error writing temporary file "{0}": {1}'.
                    format(tmpfile, exc))

        # Create bak file and then rename tmp file
        try:
            if os.path.isfile(self._connections_file):
                bakfile = '{0}.{1}'.format(self._connections_file,
                                           BAK_FILE_SUFFIX)
                if os.path.isfile(bakfile):
                    os.remove(bakfile)
                if os.path.isfile(self._connections_file):
                    os.rename(self._connections_file, bakfile)
        except (OSError, IOError) as exc:
            raise ConnectionsFileWriteError(
                self._connections_file,
                'Error renaming connections file "{0}" to backup file "{1}": '
                '{2}'.format(tmpfile, bakfile, exc))

        try:
            if self._pywbemcli_servers:
                os.rename(tmpfile, self._connections_file)
        except (OSError, IOError) as exc:
            raise ConnectionsFileWriteError(
                self._connections_file,
                'Error renaming temporary file "{0}" to connections file '
                '"{1}": {2}'.format(tmpfile, self._connections_file, exc))
        if self._verbose:
            click.echo("Connections file saved: {}".
                       format(self._connections_file))
